﻿    include 'mkl_vsl.f90'
    
    !**************************************************************************
    ! This code is written for "Smoking Initiation -- Peers and Personality",
    ! Chih-Sheng Hsieh & Hans van Kippersluis
    ! Quantitative Economics
    !
    ! Replication of Table 4: Peer Eﬀects on Smoking Dummy and Frequency 
    !– SAR models with both Networks and Personalities assumed Exogenous
    !
    ! * users should change working directory in path1, path2, and path3
    ! * GG denotes the number of school networks
    ! * TT denotes the length of MCMC iteractions
    ! * MAX denotes the maximum size of school networks
    ! * DD denotes the dimension of latent variables
    ! * THIN denotes the thinning of the MCMC chain
    ! * Nthreads denotes the CPU threads for computing
    ! * H denotes the block size of updating latent variables
    ! * dep=2: smoking dummy
    ! * dep=3: smoking frequency
    ! * raw_norm=0: raw unnormalized weights matrix
    ! * raw_norm=1: raw normalized weights matrix
    ! * burn denotes the length of burn-in iterations
    !
    !
    ! Please address any questions to:    
    ! Chih-Sheng Hsieh
    ! cshshieh@cuhk.edu.hk
    !***************************************************************************

    PROGRAM estimation

    USE random
    USE TOOL
    USE lapack95
    USE MKL_VSL_TYPE
    USE MKL_VSL
    USE omp_lib
    IMPLICIT NONE


    !****************** GENERAL VARIABLES ****************!
    CHARACTER(100) :: path1='D:/research/2014_health_study/'
    CHARACTER(100) :: path2='/data/inhome/smoke/neuroticism/'
    CHARACTER(100) :: path3='/result/SAR/smoke_frequency/homo/'
    INTEGER, PARAMETER :: GG=118, TT=50000, MAX=200, dep=3, THIN=1000, Nthreads=8, H=3, raw_norm=0, burn=100000
    REAL, PARAMETER    :: pi=3.14159265
    REAL               :: c_1, c_2, acc_1, acc_2, rand, c_3(GG), acc_3(GG),  acc_4, acc_3v , c_0(GG), acc_0(GG), kk, AICM
    INTEGER            :: NN, N(GG),  k, flag(GG)
    INTEGER            :: brng, method, me, i, j, g, t, d, ier,  r, seedsize, seed(2), errorflag, INFO, seed_mkl, t_round
    INTEGER            :: clock_max, clock_rate, clock_start, clock_stop, ithread, iblock,ibegin,iend
    REAL, DIMENSION(MAX,MAX,GG):: grade, sex, race, network
    REAL, DIMENSION(MAX,18,GG):: group
    REAL, DIMENSION(MAX,MAX,GG) :: W_norm
    REAL, ALLOCATABLE:: W(:,:), IPIV(:)
    REAL, ALLOCATABLE:: Y(:), X(:,:), rowsum_w_norm(:)
    INTEGER, ALLOCATABLE :: order(:),  INDX(:)
    LOGICAL            :: first=.true.
    TYPE (VSL_STREAM_STATE) :: stream
    INTEGER ::  errcode
    INTEGER :: CODE, SEED_PERM
    REAL(kind=8)    :: pp

    !*** posterior draws of parameters *******!

    REAL, DIMENSION(TT)       :: lambda_T
    REAL, DIMENSION(TT)       :: acc_rate1, acc_rate2, acc_rate4
    REAL, DIMENSION(24, TT)   :: beta_T
    REAL, DIMENSION(TT)       :: sigma_T
    REAL, DIMENSION(GG,TT)    :: alpha_T, acc_rate3
    REAL, DIMENSION(GG)       :: acc_rate0


    !*** hyperparameters in the prior distribution ***!
    REAL                    :: lambda_1, lambda_2, lambda_0, COV_lambda_1
    REAL, DIMENSION(24)     :: beta_0
    REAL, DIMENSION(24,24)  :: B_0, INV_B_0
    REAL                    :: eta_0, ALPHA_0, a_0
    INTEGER                 :: rho_0

    !******  VARIABLE IN M-H FOR Z_G  *******************************!

    REAL, ALLOCATABLE  :: V(:,:), S1(:,:), S2(:,:), XX(:,:)
    REAL, ALLOCATABLE  :: V1(:,:), V2(:,:)
    REAL, ALLOCATABLE  :: zero(:)
    REAL, ALLOCATABLE  :: ep(:), ep_1(:), ep_2(:)
    REAL, ALLOCATABLE  :: ep_1v(:), ep_2v(:)
    REAL(kind=8) ::   q_1, q_2, like_Y1, like_Y2, pp_w, q

    !REAL, DIMENSION(MAX,DD,GG,TT/THIN):: zz_T

    REAL(kind=8)  :: pp_l(GG), pp_G(GG)

    !*************VARIABLE IN GIBBS FOR BETA **************!

    REAL :: XVX(24,24,GG), XVY(24,GG), INV_B(24,24), INV_B2(24,24), beta_temp(24)
    REAL, ALLOCATABLE ::  XX2(:,:), YY(:), S(:,:)

    !*************VARIABLE IN M-H FOR SIGMA_E & SIGMA_EZ & ALPHA_G**************!

    REAL              :: sigma_1, sigma_2
    REAL              :: R_g,ep_v
    INTEGER           :: rho_1, stack

    !****************************************************************************!

100 format (14f10.4)
200 format (4f10.4)

    seedsize=2
    seed(1)=2.1103e+005
    seed(2)=2.1103e+005
    CALL RANDOM_SEED
    CALL RANDOM_SEED(SIZE=seedsize)
    CALL RANDOM_SEED(PUT=seed)

    seed_mkl=12345
    brng=VSL_BRNG_MCG31
    method=VSL_RNG_METHOD_GAUSSIANMV_BOXMULLER
    me=VSL_MATRIX_STORAGE_FULL

    errcode=vslnewstream(stream,brng,seed_mkl)

    CALL networksize(GG, MAX, path1, path2, N)
    CALL define_variable(GG, grade, sex, race, network, group, N, path1, path2, MAX)

    W_norm=0.0
    lambda_0=0.0
    beta_0=0.0
    B_0=eye(24)*10.0
    CALL FINDInv(B_0,INV_B_0,24,ier)
    eta_0=0.1
    rho_0=2
    ALPHA_0=400.0
    a_0=0.0


    acc_2=0.0
    acc_rate2=0.0

    ! initial value of the draws !
    sigma_T(1)=1.0


    OPEN (1000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'lambda.txt', status='unknown')
    OPEN (2000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'beta.txt', status='unknown')
    OPEN (3000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'sigma.txt', status='unknown')
    OPEN (4000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'alpha.txt', status='unknown')


    ! read raw data and sort by gender -> four spatial weight matrix !

    DO g=1,GG
        NN=N(g)
        ALLOCATE(W(NN,NN));
        ALLOCATE(rowsum_w_norm(NN))

        W=network(1:NN,1:NN,g)
        w_norm(1:NN,1:NN,g)=W

        if (raw_norm==1) THEN
            rowsum_w_norm=sum(W, dim=2)
            DO i=1,NN
                if (rowsum_w_norm(i)==0) rowsum_w_norm(i)=1;
                w_norm(i,1:NN,g)=W(i,1:NN)/rowsum_w_norm(i);

            END DO
        END IF
        DEALLOCATE(W,rowsum_w_norm)

    END DO

    DO t=2, TT

        call system_clock ( clock_start, clock_rate, clock_max )

        DO
            lambda_1=random_normal()*0.01+lambda_T(t-1)
            IF (t>2) THEN
                lambda_2=random_normal()*sqrt(var(lambda_T(1:t-1),t-1))*2.38+lambda_T(t-1)
                lambda_1=0.95*lambda_2+0.05*lambda_1
            END IF
            IF ((lambda_1>=-0.10) .and. (lambda_1<=0.10)) EXIT
        END DO

        Call OMP_SET_NUM_THREADS(Nthreads)
        !$omp parallel default(shared) private(g, NN,V, S, Y,X, XX, W, &
        !$omp & ep, ep_1, ep_2, S1, S2, INFO, CODE, &
        !$omp & IPIV, INDX, q_1, q_2, &
        !$omp & zero, pp, like_Y1, like_Y2, rand, &
        !$omp & ithread, iblock, ibegin, iend)

        ithread = OMP_GET_THREAD_NUM()
        iblock = (GG+Nthreads-1)/Nthreads
        ibegin = ithread*iblock+1
        iend = min((ithread+1)*iblock,GG)

        DO g = ibegin,iend

            NN=N(g)
            ALLOCATE(V(NN,NN));        ALLOCATE(S(NN,NN));
            ALLOCATE(S1(NN,NN));       ALLOCATE(S2(NN,NN));
            ALLOCATE(Y(NN));
            ALLOCATE(X(NN,12)) ;       ALLOCATE(XX(NN,24)) ;
            ALLOCATE(ep_1(NN));        ALLOCATE(ep_2(NN));
            ALLOCATE(IPIV(NN));        ALLOCATE(INDX(NN));

            V=sigma_T(t-1)*EYE(NN)
            Y=group(1:NN,dep,g)
            X=group(1:NN,7:18,g)
            XX(:,1:12)=X
            XX(:,13:24)=matmul(w_norm(1:NN,1:NN,g),X)

            S1=eye(NN)-lambda_1*w_norm(1:NN,1:NN,g)
            S2=eye(NN)-lambda_T(t-1)*w_norm(1:NN,1:NN,g)

            ep_1=matmul(S1,Y)-matmul(XX,beta_T(:,t-1))-alpha_T(g,t-1)
            ep_2=matmul(S2,Y)-matmul(XX,beta_T(:,t-1))-alpha_T(g,t-1)

            like_Y1=log(FindDet(S1,NN))-dot_product(ep_1, ep_1)/(2*sigma_T(t-1))
            like_Y2=log(FindDet(S2,NN))-dot_product(ep_2, ep_2)/(2*sigma_T(t-1))

            pp_l(g)=(like_Y1-like_Y2)

            DEALLOCATE(V);        DEALLOCATE(S);
            DEALLOCATE(S1);       DEALLOCATE(S2);
            DEALLOCATE(Y);
            DEALLOCATE(X) ;       DEALLOCATE(XX) ;
            DEALLOCATE(ep_1);     DEALLOCATE(ep_2);
            DEALLOCATE(IPIV);     DEALLOCATE(INDX);

        END DO
        !$omp end parallel

        CALL random_number(rand)
        IF (log(rand)<=sum(pp_l)) THEN
            lambda_T(t)=lambda_1
            acc_2=acc_2+1.0
        ELSE
            lambda_T(t)=lambda_T(t-1)
        END IF
        acc_rate2(t)=acc_2/t

        !***************GIBBS STEP TO SIMULTE BETA *********************!

        XVX=0.0
        XVY=0.0

        Call OMP_SET_NUM_THREADS(Nthreads)
        !$omp parallel default(shared) private(g,NN, V, Y, YY, &
        !$omp & X, XX, XX2, W, S, V1, V2, IPIV, INFO, &
        !$omp & ithread, iblock, ibegin, iend)

        ithread = OMP_GET_THREAD_NUM()
        iblock = (GG+Nthreads-1)/Nthreads
        ibegin = ithread*iblock+1
        iend = min((ithread+1)*iblock,GG)

        DO g = ibegin,iend

            NN=N(g)
            ALLOCATE(V(NN,NN))
            ALLOCATE(Y(NN));        ALLOCATE(YY(NN))
            ALLOCATE(X(NN,12)) ;
            ALLOCATE(XX(NN,24)) ; ALLOCATE(XX2(NN,24)) ;
            ALLOCATE(S(NN,NN));
            ALLOCATE(V1(NN,NN)); ALLOCATE(V2(NN,NN))
            ALLOCATE(IPIV(NN))

            V=sigma_T(t-1)*EYE(NN)
            Y=group(1:NN,dep,g)
            X=group(1:NN,7:18,g)
            XX(:,1:12)=X
            XX(:,13:24)=matmul(w_norm(1:NN,1:NN,g),X)

            S=eye(NN)-lambda_T(t)*w_norm(1:NN,1:NN,g)

            XX2=XX
            V1=V
            CALL SGESV(NN,24,V1,NN,IPIV,XX2,NN,INFO)
            YY=matmul(S,Y)-alpha_T(g,t-1)
            V2=V
            CALL SGESV(NN,1,V2,NN,IPIV,YY,NN,INFO)

            XVX(:,:,g)=matmul(transpose(XX) , XX2)
            XVY(:,g)=matmul(transpose(XX) , YY)

            DEALLOCATE(V,Y,YY,X,XX,XX2,S,V1,V2,IPIV)
        END DO
        !$omp end parallel

        CALL FINDInv(sum(XVX,3)+INV_B_0,INV_B,24,ier)
        INV_B2=INV_B
        CALL spotrf('U',24, INV_B2, 24, info)
        errcode=vsrnggaussianmv(method, stream, 1, beta_temp, 24, me, beta_0,  INV_B2)
        beta_T(:,t)=beta_temp+matmul(INV_B,sum(XVY,2))


        !******************M-H STEP FOR SIGMA_E *****************!

        ep_v=0.0
        rho_1=0
        stack=0
        DO g=1,GG
            NN=N(G)
            ALLOCATE(Y(NN))
            ALLOCATE(X(NN,12))
            ALLOCATE(XX(NN,24))
            ALLOCATE(S(NN,NN))
            ALLOCATE(ep(NN))

            Y=group(1:NN,dep,g)
            X=group(1:NN,7:18,g)
            XX(:,1:12)=X
            XX(:,13:24)=matmul(w_norm(1:NN,1:NN,g),X)

            S=eye(NN)-lambda_T(t)*w_norm(1:NN,1:NN,g)

            ep=matmul(S,Y)-matmul(XX,beta_T(:,t))-alpha_T(g,t-1)
            ep_v=ep_v+dot_product(ep,ep)
            rho_1=rho_1+NN
            stack=stack+NN
            DEALLOCATE(Y,X,XX,S,ep)
        END DO
        sigma_T(t)=(ep_v+eta_0)/random_chisq(rho_1+rho_0, first)



        !*************GIBBS STEP FOR ALPHA_G*********************!


        Call OMP_SET_NUM_THREADS(Nthreads)
        !$omp parallel default(shared) private(g,NN, Y, YY, &
        !$omp & X, XX, S, IPIV, R_g, &
        !$omp & ithread, iblock, ibegin, iend)

        ithread = OMP_GET_THREAD_NUM()
        iblock = (GG+Nthreads-1)/Nthreads
        ibegin = ithread*iblock+1
        iend = min((ithread+1)*iblock,GG)

        DO g = ibegin,iend

            NN=N(g)
            ALLOCATE(Y(NN));
            ALLOCATE(YY(NN));
            ALLOCATE(X(NN,12)) ;
            ALLOCATE(XX(NN,24)) ;
            ALLOCATE(S(NN,NN));

            R_g=(ALPHA_0**(-1)+sigma_T(t)**(-1.0)*NN)**(-1.0)
            Y=group(1:NN,dep,g)
            X=group(1:NN,7:18,g)
            XX(:,1:12)=X
            XX(:,13:24)=matmul(w_norm(1:NN,1:NN,g),X)

            S=eye(NN)-lambda_T(t)*w_norm(1:NN,1:NN,g)

            YY=matmul(S,Y)
            alpha_T(g,t)=R_g*(ALPHA_0**(-1)*a_0+(sigma_T(t))**(-1.0)*sum((YY-matmul(XX,beta_T(:,t)))))+random_normal()*sqrt(R_g)

            DEALLOCATE(Y,YY,X,XX,S)
        END DO
        !$omp end parallel


        call system_clock ( clock_stop, clock_rate, clock_max )
        t_round=t/20
        IF (t-t_round*20==0) THEN
            WRITE(*,*)   'T=', t
            WRITE(*,'(a6, f7.4)') 'Time:', real ( clock_stop - clock_start ) / real ( clock_rate )
            WRITE(*,'(a6,1f7.4)') 'LAMBDA:', lambda_T(t)
            WRITE(*,'(a6,8f7.4)') 'BETA:', beta_T(1:8,t)
            WRITE(*,'(a6,8f7.4)') 'BETA:', beta_T(9:15,t)
            WRITE(*,'(a6,7f7.4)') 'SIGMA:',  sigma_T(t)

            WRITE(1000,'(1f10.4)')   lambda_T(t)
            WRITE(2000,'(30f10.4)')  beta_T(:,t)
            WRITE(3000,'(f12.3)')    sigma_T(t)
            WRITE(4000,100)          alpha_T(:,t)
        END IF

        !t_round=t/THIN
        !IF (t-t_round*THIN==0) THEN
        !    WRITE(5000,200)  ((zz(i,:,g),i=1,300),g=1,GG)
        !END IF


    END DO



    END PROGRAM estimation



