    INCLUDE 'mkl_vsl.f90'

    !**************************************************************************
    ! This code is written for "Smoking Initiation -- Peers and Personality",
    ! Chih-Sheng Hsieh & Hans van Kippersluis
    ! Quantitative Economics
    !
    ! Replication of Table A10: Placebo Test for Examining Peer effects
    ! on Father's Education
    !
    ! * users should change working directory in path1, path2, and path3
    ! * GG denotes the number of school networks
    ! * TT denotes the length of MCMC iteractions
    ! * MAX denotes the maximum size of school networks
    ! * DD denotes the dimension of latent variables
    ! * THIN denotes the thinning of the MCMC chain
    ! * Nthreads denotes the CPU threads for computing
    ! * H denotes the block size of updating latent variables
    ! * dep=2: smoking dummy
    ! * dep=3: smoking frequency
    ! * raw_norm=0: raw unnormalized weights matrix
    ! * raw_norm=1: raw normalized weights matrix
    ! * burn denotes the length of burn-in iterations
    ! * users should change initial gamma values w.r.t. DD in line 169-172
    !
    ! Please address any questions to:    
    ! Chih-Sheng Hsieh
    ! cshshieh@cuhk.edu.hk
    !***************************************************************************

    PROGRAM estimation

    USE random
    USE TOOL
    USE lapack95
    USE MKL_VSL_TYPE
    USE MKL_VSL
    USE omp_lib
    IMPLICIT NONE

    !****************** GENERAL VARIABLES ****************!
    CHARACTER(100) :: path1='D:/research/2014_health_study/'
    CHARACTER(100) :: path2='/data_inhome/dad_edu/'
    CHARACTER(100) :: path3='/result/result_inhome_dadedu/four_dim_new/'
    CHARACTER(40) :: numchr
    INTEGER, PARAMETER :: GG=118, TT=170000, MAX=200, DD=4, THIN=1000, Nthreads=120, H=2, raw_norm=0
    REAL, PARAMETER    :: pi=3.14119265
    REAL               :: c_1, c_2, c_4, acc_1, acc_2, rand, c_3(GG), acc_3(GG),  acc_4, acc_3v , c_0(GG), acc_0(GG), kk
    INTEGER            :: NN, N(GG),  k, censor(GG), flag(GG)
    INTEGER            :: brng, method, me, i, j, g, t, d, ier,  r, seedsize, seed(2), errorflag, INFO, seed_mkl, t_round
    INTEGER            :: clock_max, clock_rate, clock_start, clock_stop, ithread, iblock,ibegin,iend
    REAL, DIMENSION(MAX,MAX,GG):: grade, sex, race, network
    REAL, DIMENSION(MAX,12,GG):: group
    REAL, DIMENSION(MAX,MAX,GG) :: W_norm, W_FF, W_FM, W_MF, W_MM
    REAL, ALLOCATABLE:: C1(:,:), C2(:,:), C3(:,:), W(:,:), IPIV(:)
    REAL, ALLOCATABLE:: Y(:), X(:,:), personality(:)
    INTEGER, ALLOCATABLE :: order(:),  INDX(:)
    LOGICAL            :: first=.true.
    TYPE (VSL_STREAM_STATE) :: stream
    INTEGER ::  errcode
    INTEGER :: CODE, DD1, DD2, SEED_PERM
    REAL(kind=8)    :: det1, det2, pp


    !*** hyperparameters in the prior distribution ***!
    REAL, DIMENSION(4+DD)      :: gamma_0
    REAL, DIMENSION(22)        :: beta_0
    REAL, DIMENSION(4+DD,4+DD) :: G_0
    REAL, DIMENSION(22,22)     :: B_0, INV_B_0
    REAL, DIMENSION(1+DD)      :: sigma_0
    REAL, DIMENSION(1+DD,1+DD) :: Sig_0
    REAL                       ::  ALPHA_0, a_0


    !*** posterior draws of parameters *******!

    REAL, DIMENSION(4+DD,TT)  :: gamma_T
    REAL, DIMENSION(TT)    :: lambda_T
    REAL, DIMENSION(TT)       :: acc_rate1, acc_rate2, acc_rate4
    REAL(kind=8), DIMENSION(TT) :: logp_T
    REAL, DIMENSION(22, TT)   :: beta_T
    REAL, DIMENSION(DD+1,TT)  :: sigma_T
    REAL, DIMENSION(GG,TT)    :: alpha_T, acc_rate3
    REAL, DIMENSION(GG)       :: acc_rate0

    !******  VARIABLE IN M-H FOR Z_G  *******************************!

    REAL, DIMENSION(MAX,DD,GG):: zz
    REAL, ALLOCATABLE  :: zz_1(:,:), zz_2(:,:), V(:,:), S1(:,:), S2(:,:), XX(:,:),  cov_zz(:,:), temp_zz(:)
    REAL, ALLOCATABLE  :: V1(:,:), V2(:,:)
    REAL, ALLOCATABLE  :: psi(:,:), psi_1(:,:), psi_2(:,:), zero(:)
    REAL, ALLOCATABLE  :: ep(:), ep_1(:), ep_2(:)
    REAL, ALLOCATABLE  :: ep_1v(:), ep_2v(:)
    REAL(kind=8) ::   q_1, q_2, like_Y1, like_Y2, pp_w, q

    !REAL, DIMENSION(MAX,DD,GG,TT/THIN):: zz_T

    !********VARIABLE IN M-H FOR GAMMA & LAMBDA   ***************!

    REAL      :: lambda_1, lambda_2, COV_lambda_1, lambda_0
    REAL      :: gamma_1(4+DD), gamma_2(4+DD), COV_gamma_1(4+DD,4+DD)
    REAL(kind=8)  :: pp_l(GG), pp_G(GG), pp_sig(GG)

    !*************VARIABLE IN M-H FOR BETA **************!

    REAL :: XVX(22,22,GG), XVY(22,GG), INV_B(22,22), INV_B2(22,22), beta_temp(22)
    REAL, ALLOCATABLE ::  XX2(:,:), YY(:), S(:,:)

    !*************VARIABLE IN M-H FOR SIGMA_E & SIGMA_EZ & ALPHA_G**************!

    REAL              :: sigma_1(1+DD), sigma_2(1+DD), COV_sigma_1(1+DD,1+DD)
    REAL              ::  R_g
    REAL, ALLOCATABLE :: VV1(:,:), VV2(:,:)

    !*******************************************************************************************************!

    INTEGER :: N_1, N_2, T_1, T_2
    REAL, ALLOCATABLE:: rowsum_w_ff(:), rowsum_w_fm(:), rowsum_w_mf(:), rowsum_w_mm(:), rowsum_w_norm(:)

    !*******************************************************************************************************!

100 format (14f10.4)
200 format (4f10.4)

    seedsize=2
    seed(1)=2.1103e+005
    seed(2)=2.1103e+005
    CALL RANDOM_SEED
    CALL RANDOM_SEED(SIZE=seedsize)
    CALL RANDOM_SEED(PUT=seed)

    seed_mkl=12345
    brng=VSL_BRNG_MCG31
    method=VSL_RNG_METHOD_GAUSSIANMV_BOXMULLER
    me=VSL_MATRIX_STORAGE_FULL

    errcode=vslnewstream(stream,brng,seed_mkl)

    CALL networksize(GG, MAX, path1, path2, N)
    CALL define_variable(GG, grade, sex, race, network, group, N, path1, path2, MAX)

    zz=0.0
    lambda_0=0.0
    gamma_0=0.0
    beta_0=0.0
    G_0=eye(4+DD)*100.0
    Sig_0=eye(1+DD)*100.0
    B_0=eye(22)*100.0
    CALL FINDInv(B_0,INV_B_0,22,ier)
    zz=0.0
    ALPHA_0=400.0
    a_0=0.0
    sigma_0(1)=1.0
    sigma_0(2:DD+1)=0.5
    censor=0

    c_0=0.1
    c_3=0.10
    c_4=0.0001
    acc_1=0.0
    acc_2=0.0
    acc_3=0.0
    acc_rate1=0.0
    acc_rate2=0.0
    acc_rate3=0.0


    ! initial value of the draws !
    !gamma_T(:,1)=(/  -3.5037,  2.3110   , 0.3811  ,  0.8758 ,  8.1173/)   ! D=1 !
    !gamma_T(:,1)=(/   -1.5175  ,  2.4617 ,   0.3522  ,  0.9279 ,  5.7569  ,  5.3701 /)  ! D=2 !
    !gamma_T(:,1)=(/  -0.8025 ,   2.1720  ,  0.3167  ,  0.5420  ,  4.5590  ,  4.4948  ,  4.4364/) !D=3!
    gamma_T(:,1)=(/   0.6812  ,  2.2737  ,  0.3259  ,  0.5967  ,  3.62  ,  3.2509  ,  3.022 ,   2.7210/) !D=4!
    sigma_T(1,1)=0.1500
    sigma_T(2:DD+1,1)=0.1500


    OPEN (1000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'gamma.txt', status='unknown')
    OPEN (2000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'lambda.txt', status='unknown')
    OPEN (3000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'beta.txt', status='unknown')
    OPEN (4000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'sigma.txt', status='unknown')
    OPEN (5000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'latent_z.txt', status='replace')
    OPEN (6000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'alpha.txt', status='unknown')
    OPEN (7000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'loglikeli.txt', status='unknown')


    DO t=2, TT

        call system_clock ( clock_start, clock_rate, clock_max )

        flag=0
        DO
            lambda_1=random_normal()*0.1+lambda_T(t-1)
            IF (t>2) THEN
                lambda_2=random_normal()*sqrt(var(lambda_T(1:t-1),t-1))*2.38+lambda_T(t-1)
                lambda_1=0.95*lambda_2+0.05*lambda_1
            END IF
            IF ((lambda_1>=-0.10) .and. (lambda_1<=0.10)) EXIT
        END DO

        DO
            COV_gamma_1=EYE(4+DD)*0.01/(4+DD)
            call spotrf('U',4+DD, COV_gamma_1, 4+DD, info)
            errcode=vsrnggaussianmv(method,stream,1,gamma_1, 4+DD, me, gamma_0, COV_gamma_1)
            gamma_1=gamma_1+gamma_T(:,t-1)
            IF (t>10000) THEN
                COV_gamma_1=cov(gamma_T(:,9000:t-1),4+DD,t-9000)*2.38**2/(4+DD)
                call spotrf('U',4+DD, COV_gamma_1, 4+DD, info)
                errcode=vsrnggaussianmv(method,stream,1,gamma_2, 4+DD, me, gamma_0, COV_gamma_1)
                gamma_2=gamma_2+gamma_T(:,t-1)
                gamma_1=0.95*gamma_2+0.05*gamma_1
            END IF
            IF (DD==1) THEN
                IF (gamma_1(4+DD)>=0) EXIT
            ELSEIF (DD==2) THEN
                IF (gamma_1(3+DD)>=gamma_1(4+DD)) EXIT
            ELSEIF (DD==3) THEN
                IF ((gamma_1(2+DD)>=gamma_1(3+DD)) .and. (gamma_1(3+DD)>=gamma_1(4+DD)) ) EXIT
            ELSEIF (DD==4)  THEN
                IF ((gamma_1(1+DD)>=gamma_1(2+DD)) .and. (gamma_1(2+DD)>=gamma_1(3+DD)) .and. (gamma_1(3+DD)>=gamma_1(4+DD))) EXIT
            END IF
        END DO

        gamma_2=gamma_T(:,t-1)

        ! THE M-H ALGORITHM FOR SAMPLING Z !


        Call OMP_SET_NUM_THREADS(Nthreads)
        !$omp parallel default(shared) private(g, NN,V, S, Y,X, XX, W,&
        !$omp & psi_1, psi_2, ep, ep_1, ep_2, ep_1v, ep_2v, S1, S2, INFO, CODE, DD1, DD2, &
        !$omp & C1, C2, C3, V1, V2, zz_1, zz_2, IPIV, INDX, det1, det2, q_1, q_2, &
        !$omp & kk, acc_3v, k, cov_zz, temp_zz, zero, d, pp, i, j, like_Y1, like_Y2, rand, &
        !$omp & ithread, iblock, ibegin, iend)

        ithread = OMP_GET_THREAD_NUM()
        iblock = (GG+Nthreads-1)/Nthreads
        ibegin = ithread*iblock+1
        iend = min((ithread+1)*iblock,GG)

        DO g = ibegin,iend

            NN=N(g)
            ALLOCATE(V(NN,NN));        ALLOCATE(S(NN,NN));
            ALLOCATE(S1(NN,NN));       ALLOCATE(S2(NN,NN));
            ALLOCATE(Y(NN));
            ALLOCATE(X(NN,11)) ;       ALLOCATE(XX(NN,22)) ;
            ALLOCATE(W(NN,NN));
            ALLOCATE(psi_1(NN,NN));    ALLOCATE(psi_2(NN,NN))
            ALLOCATE(ep(NN));
            ALLOCATE(ep_1(NN));        ALLOCATE(ep_2(NN));
            ALLOCATE(ep_1v(NN));       ALLOCATE(ep_2v(NN))  ;
            ALLOCATE(C1(NN,NN));       ALLOCATE(C2(NN,NN));
            ALLOCATE(C3(NN,NN));
            ALLOCATE(V1(NN,NN));       ALLOCATE(V2(NN,NN))  ;
            ALLOCATE(zz_1(NN,DD));     ALLOCATE(zz_2(NN,DD))
            ALLOCATE(IPIV(NN));        ALLOCATE(INDX(NN));

            V=(sigma_T(1,t-1)-sum(sigma_T(2:DD+1,t-1)*sigma_T(2:DD+1,t-1)))*EYE(NN)
            Y=group(1:NN,1,g)
            X=group(1:NN,2:12,g)
            W=network(1:NN,1:NN,g)
            XX(:,1:11)=X
            XX(:,12:22)=matmul(W,X)
            zz_1=zz(1:NN,1:DD,g)
            zz_2=zz(1:NN,1:DD,g)

            C1=grade(1:NN,1:NN,g)
            C2=sex(1:NN,1:NN,g)
            C3=race(1:NN,1:NN,g)

            S=eye(NN)-lambda_T(t-1)*W

            ep=matmul(S,Y)-matmul(XX,beta_T(:,t-1))-alpha_T(g,t-1)

            kk=NN
            kk=floor(kk/H)
            IF (kk==0) kk=kk+1

            acc_3v=0.0
            DO k=1, kk
                IF(k==kk) THEN
                    ALLOCATE(cov_zz(NN-(k-1)*H,NN-(k-1)*H))
                    ALLOCATE(temp_zz(NN-(k-1)*H))
                    ALLOCATE(zero(NN-(k-1)*H))
                    zero=0.0
                    cov_zz=EYE(NN-(k-1)*H)*c_3(g)
                    call spotrf('U',NN-(k-1)*H, cov_zz, NN-(k-1)*H, info)
                    DO d=1, DD
                        errcode=vsrnggaussianmv(method,stream, 1, temp_zz, NN-(k-1)*H, me, zero, cov_zz)
                        zz_1((k-1)*H+1:NN,d)=temp_zz+zz_2((k-1)*H+1:NN,d)
                    END DO
                    DEALLOCATE(cov_zz)
                    DEALLOCATE(temp_zz)
                    DEALLOCATE(zero)
                ELSE
                    ALLOCATE(cov_zz(H,H))
                    ALLOCATE(temp_zz(H))
                    ALLOCATE(zero(H))
                    zero=0.0
                    cov_zz=EYE(H)*c_3(g)
                    call spotrf('U',H, cov_zz, H, info)
                    DO d=1,DD
                        errcode=vsrnggaussianmv(method,stream, 1, temp_zz, H, me, zero, cov_zz)
                        zz_1((k-1)*H+1:k*H,d)=temp_zz+zz_2((k-1)*H+1:k*H,d)
                    END DO
                    DEALLOCATE(cov_zz)
                    DEALLOCATE(temp_zz)
                    DEALLOCATE(zero)
                END IF

                pp=0.0
                DO i=1,NN
                    DO j=1,NN
                        IF (j/=i) THEN
                            IF (DD==1) THEN
                                psi_1(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j) &
                                    & -gamma_T(5,t-1)*abs(zz_1(i,1)-zz_1(j,1))
                                psi_2(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j) &
                                    & -gamma_T(5,t-1)*abs(zz_2(i,1) -zz_2(j,1))
                            ELSEIF (DD==2) THEN
                                psi_1(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j) &
                                    & -gamma_T(5,t-1)*abs(zz_1(i,1)-zz_1(j,1))-gamma_T(6,t-1)*abs(zz_1(i,2)-zz_1(j,2))
                                psi_2(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j) &
                                    & -gamma_T(5,t-1)*abs(zz_2(i,1)-zz_2(j,1))-gamma_T(6,t-1)*abs(zz_2(i,2)-zz_2(j,2))
                            ELSEIF (DD==3) THEN
                                psi_1(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j)  &
                                    & -gamma_T(5,t-1)*abs(zz_1(i,1)-zz_1(j,1)) -gamma_T(6,t-1)*abs(zz_1(i,2)-zz_1(j,2)) &
                                    & -gamma_T(7,t-1)*abs(zz_1(i,3)-zz_1(j,3))
                                psi_2(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j)  &
                                    & -gamma_T(5,t-1)*abs(zz_2(i,1)-zz_2(j,1)) -gamma_T(6,t-1)*abs(zz_2(i,2)-zz_2(j,2)) &
                                    & -gamma_T(7,t-1)*abs(zz_2(i,3)-zz_2(j,3))
                            ELSEIF (DD==4) THEN
                                psi_1(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j)  &
                                    & -gamma_T(5,t-1)*abs(zz_1(i,1)-zz_1(j,1))-gamma_T(6,t-1)*abs(zz_1(i,2)-zz_1(j,2)) &
                                    & -gamma_T(7,t-1)*abs(zz_1(i,3)-zz_1(j,3))-gamma_T(8,t-1)*abs(zz_1(i,4)-zz_1(j,4))
                                psi_2(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j)  &
                                    & -gamma_T(5,t-1)*abs(zz_2(i,1)-zz_2(j,1)) -gamma_T(6,t-1)*abs(zz_2(i,2)-zz_2(j,2))  &
                                    & -gamma_T(7,t-1)*abs(zz_2(i,3)-zz_2(j,3)) -gamma_T(8,t-1)*abs(zz_2(i,4)-zz_2(j,4))
                            ENDIF

                            q_1=psi_1(i,j)*W(i,j)-log(1+exp(psi_1(i,j)))
                            q_2=psi_2(i,j)*W(i,j)-log(1+exp(psi_2(i,j)))
                            pp=pp+(q_1-q_2)
                        ENDIF
                    END DO
                END DO


                ep_1=ep-matmul(zz_1,sigma_T(2:DD+1,t-1))
                ep_2=ep-matmul(zz_2,sigma_T(2:DD+1,t-1))
                ep_1v=ep_1; ep_2v=ep_2
                V1=V; V2=V

                CALL SGESV(NN,1,V1,NN,IPIV,ep_1v,NN,INFO)
                CALL SGESV(NN,1,V2,NN,IPIV,ep_2v,NN,INFO)

                like_Y1=-dot_product(ep_1, ep_1v)/2
                like_Y2=-dot_product(ep_2, ep_2v)/2

                IF (k==kk) THEN
                    pp=pp+(like_Y1-like_Y2)+logmvnpdf(reshape(zz_1((k-1)*H+1:NN,:),(/(NN-(k-1)*H)*DD,1/)),eye((NN-(k-1)*H)*DD),(NN-(k-1)*H)*DD)-logmvnpdf(reshape(zz_2((k-1)*H+1:NN,:),(/(NN-(k-1)*H)*DD,1/)),eye((NN-(k-1)*H)*DD),(NN-(k-1)*H)*DD)
                ELSE
                    pp=pp+(like_Y1-like_Y2)+logmvnpdf(reshape(zz_1((k-1)*H+1:k*H,:),(/H*DD,1/)),eye(H*DD),H*DD)-logmvnpdf(reshape(zz_2((k-1)*H+1:k*H,:),(/H*DD,1/)),eye(H*DD),H*DD)
                END IF

                CALL random_number(rand)
                IF (log(rand)<=pp) THEN
                    zz_2=zz_1
                    acc_3v=acc_3v+1
                END IF
                zz_1=zz_2
            END DO
            if (3*acc_3v>1*kk) acc_3(g)=acc_3(g)+1
            acc_rate3(g,t)=acc_3(g)/t
            if ((acc_rate3(g,t)<0.05)) c_3(g)=c_3(g)/1.01
            if ((acc_rate3(g,t)>0.15)) c_3(g)=c_3(g)*1.01
            zz(1:NN,1:DD,g)=zz_1
            !t_round=t/THIN
            !IF (t-t_round*THIN==0) THEN
            !    zz_T(:,:,g,t/THIN)=zz(:,:,g)
            !END IF


            !***** THE M-H ALGORITHM FOR SAMPLING GAMMA & LAMBDA ******!

            pp=0.0
            DO i=1,NN
                DO j=1, NN
                    IF (j/=i) THEN
                        IF (DD==1) THEN
                            psi_1(i,j)=gamma_1(1)+gamma_1(2)*c1(i,j)+gamma_1(3)*c2(i,j)+gamma_1(4)*c3(i,j) -gamma_1(5)*abs(zz_1(i,1)-zz_1(j,1))
                            psi_2(i,j)=gamma_2(1)+gamma_2(2)*c1(i,j)+gamma_2(3)*c2(i,j)+gamma_2(4)*c3(i,j) -gamma_2(5)*abs(zz_1(i,1)-zz_1(j,1))
                        ELSEIF (DD==2) THEN
                            psi_1(i,j)=gamma_1(1)+gamma_1(2)*c1(i,j)+gamma_1(3)*c2(i,j)+gamma_1(4)*c3(i,j) -gamma_1(5)*abs(zz_1(i,1)-zz_1(j,1)) &
                                & -gamma_1(6)*abs(zz_1(i,2)-zz_1(j,2))
                            psi_2(i,j)=gamma_2(1)+gamma_2(2)*c1(i,j)+gamma_2(3)*c2(i,j)+gamma_2(4)*c3(i,j) -gamma_2(5)*abs(zz_1(i,1)-zz_1(j,1)) &
                                & -gamma_2(6)*abs(zz_1(i,2)-zz_1(j,2))
                        ELSEIF (DD==3) THEN
                            psi_1(i,j)=gamma_1(1)+gamma_1(2)*c1(i,j)+gamma_1(3)*c2(i,j)+gamma_1(4)*c3(i,j) -gamma_1(5)*abs(zz_1(i,1)-zz_1(j,1))-gamma_1(6)*abs(zz_1(i,2)-zz_1(j,2)) &
                                &-gamma_1(7)*abs(zz_1(i,3)-zz_1(j,3))
                            psi_2(i,j)=gamma_2(1)+gamma_2(2)*c1(i,j)+gamma_2(3)*c2(i,j)+gamma_2(4)*c3(i,j) -gamma_2(5)*abs(zz_1(i,1)-zz_1(j,1))-gamma_2(6)*abs(zz_1(i,2)-zz_1(j,2)) &
                                &-gamma_2(7)*abs(zz_1(i,3)-zz_1(j,3))
                        ELSEIF (DD==4) THEN
                            psi_1(i,j)=gamma_1(1)+gamma_1(2)*c1(i,j)+gamma_1(3)*c2(i,j)+gamma_1(4)*c3(i,j) -gamma_1(5)*abs(zz_1(i,1)-zz_1(j,1))-gamma_1(6)*abs(zz_1(i,2)-zz_1(j,2)) &
                                & -gamma_1(7)*abs(zz_1(i,3)-zz_1(j,3))-gamma_1(8)*abs(zz_1(i,4)-zz_1(j,4))
                            psi_2(i,j)=gamma_2(1)+gamma_2(2)*c1(i,j)+gamma_2(3)*c2(i,j)+gamma_2(4)*c3(i,j) -gamma_2(5)*abs(zz_1(i,1)-zz_1(j,1))-gamma_2(6)*abs(zz_1(i,2)-zz_1(j,2)) &
                                & -gamma_2(7)*abs(zz_1(i,3)-zz_1(j,3))-gamma_2(8)*abs(zz_1(i,4)-zz_1(j,4))
                        ENDIF
                        q_1=psi_1(i,j)*W(i,j)-log(1+exp(psi_1(i,j)))
                        q_2=psi_2(i,j)*W(i,j)-log(1+exp(psi_2(i,j)))
                        pp=pp+(q_1-q_2)
                    ENDIF
                END DO
            END DO

            pp_G(g)=pp


            S1=eye(NN)-lambda_1*W
            S2=eye(NN)-lambda_T(t-1)*W

            ep_1=matmul(S1,Y)-matmul(XX,beta_T(:,t-1))-matmul(zz_1,sigma_T(2:DD+1,t-1))-alpha_T(g,t-1)
            ep_2=matmul(S2,Y)-matmul(XX,beta_T(:,t-1))-matmul(zz_1,sigma_T(2:DD+1,t-1))-alpha_T(g,t-1)

            ep_1v=ep_1
            ep_2v=ep_2
            V1=V
            V2=V

            CALL SGESV(NN,1,V1,NN,IPIV,ep_1v,NN,INFO)
            CALL SGESV(NN,1,V2,NN,IPIV,ep_2v,NN,INFO)
            like_Y1=log(FindDet(S1, NN))-dot_product(ep_1, ep_1v)/2
            like_Y2=log(FindDet(S2, NN))-dot_product(ep_2, ep_2v)/2

            pp_l(g)=(like_Y1-like_Y2)

            DEALLOCATE(V);        DEALLOCATE(S);
            DEALLOCATE(S1);       DEALLOCATE(S2);
            DEALLOCATE(Y);
            DEALLOCATE(X) ;       DEALLOCATE(XX) ;
            DEALLOCATE(W);
            DEALLOCATE(psi_1);    DEALLOCATE(psi_2)
            DEALLOCATE(ep);
            DEALLOCATE(ep_1);     DEALLOCATE(ep_2);
            DEALLOCATE(ep_1v);    DEALLOCATE(ep_2v)  ;
            DEALLOCATE(C1);       DEALLOCATE(C2);
            DEALLOCATE(C3);
            DEALLOCATE(V1);       DEALLOCATE(V2)  ;
            DEALLOCATE(zz_1);     DEALLOCATE(zz_2)
            DEALLOCATE(IPIV);     DEALLOCATE(INDX);

        END DO
        !$omp end parallel

        pp=sum(pp_G)+logmvnpdf(gamma_1, G_0, 4+DD)-logmvnpdf(gamma_2, G_0, 4+DD)

        CALL random_number(rand)
        IF (log(rand)<=pp) THEN
            gamma_T(:,t)=gamma_1
            acc_1=acc_1+1.0
        ELSE
            gamma_T(:,t)=gamma_T(:,t-1)
        ENDIF
        acc_rate1(t)=acc_1/t

        CALL random_number(rand)
        IF (log(rand)<=sum(pp_l)) THEN
            lambda_T(t)=lambda_1
            acc_2=acc_2+1.0
        ELSE
            lambda_T(t)=lambda_T(t-1)
        END IF
        acc_rate2(t)=acc_2/t

        !***************GIBBS STEP TO SIMULTE BETA *********************!

        XVX=0.0
        XVY=0.0

        Call OMP_SET_NUM_THREADS(Nthreads)
        !$omp parallel default(shared) private(g,NN, V, Y, YY, &
        !$omp & X, XX, XX2, W, S1, V1, V2, zz_1, IPIV, INFO, &
        !$omp & ithread, iblock, ibegin, iend)

        ithread = OMP_GET_THREAD_NUM()
        iblock = (GG+Nthreads-1)/Nthreads
        ibegin = ithread*iblock+1
        iend = min((ithread+1)*iblock,GG)

        DO g = ibegin,iend

            NN=N(g)
            ALLOCATE(V(NN,NN))
            ALLOCATE(Y(NN));        ALLOCATE(YY(NN))
            ALLOCATE(X(NN,11)) ;
            ALLOCATE(XX(NN,22)) ; ALLOCATE(XX2(NN,22)) ;
            ALLOCATE(W(NN,NN));
            ALLOCATE(S1(NN,NN));
            ALLOCATE(V1(NN,NN)); ALLOCATE(V2(NN,NN))
            ALLOCATE(zz_1(NN,DD))
            ALLOCATE(IPIV(NN))

            V=(sigma_T(1,t-1)-sum(sigma_T(2:DD+1,t-1)*sigma_T(2:DD+1,t-1)))*EYE(NN)
            Y=group(1:NN,1,g)
            X=group(1:NN,2:12,g)
            W=network(1:NN,1:NN,g)
            XX(:,1:11)=X
            XX(:,12:22)=matmul(W,X)
            zz_1=zz(1:NN,1:DD,g)
            S1=eye(NN)-lambda_T(t)*W

            XX2=XX
            V1=V
            CALL SGESV(NN,22,V1,NN,IPIV,XX2,NN,INFO)
            YY=matmul(S1,Y)-matmul(zz_1,sigma_T(2:DD+1,t-1))-alpha_T(g,t-1)
            V2=V
            CALL SGESV(NN,1,V2,NN,IPIV,YY,NN,INFO)

            XVX(:,:,g)=matmul(transpose(XX) , XX2)
            XVY(:,g)=matmul(transpose(XX) , YY)

            DEALLOCATE(V)
            DEALLOCATE(Y);    DEALLOCATE(YY)
            DEALLOCATE(X) ;
            DEALLOCATE(XX) ; DEALLOCATE(XX2) ;
            DEALLOCATE(W);
            DEALLOCATE(S1);
            DEALLOCATE(V1); DEALLOCATE(V2)
            DEALLOCATE(zz_1)
            DEALLOCATE(IPIV)
        END DO
        !$omp end parallel

        CALL FINDInv(sum(XVX,3)+INV_B_0,INV_B,22,ier)
        INV_B2=INV_B
        CALL spotrf('U',22, INV_B2, 22, info)
        errcode=vsrnggaussianmv(method, stream, 1, beta_temp, 22, me, beta_0,  INV_B2)
        beta_T(:,t)=beta_temp+matmul(INV_B,sum(XVY,2))


        !******************M-H STEP FOR SIGMA_E & SIGMA_EZ *****************!


        ALLOCATE(zero(1+DD))
        zero=0.0
        DO
            COV_sigma_1=EYE(1+DD)*c_4
            call spotrf('U',1+DD, COV_sigma_1, 1+DD, info)
            errcode=vsrnggaussianmv(method,stream,1,sigma_1, 1+DD, me, zero, COV_sigma_1)
            sigma_1=sigma_1+sigma_T(:,t-1)

            IF (t>500) THEN
                COV_sigma_1=cov(sigma_T(:,1:t-1),1+DD,t-1)*2.38**2/(1+DD)
                call spotrf('U',1+DD, COV_sigma_1, 1+DD, info)
                errcode=vsrnggaussianmv(method,stream, 1, sigma_2, 1+DD, me, zero, COV_sigma_1)
                sigma_2=sigma_2+sigma_T(:,t-1)
                sigma_1=0.95*sigma_2+0.05*sigma_1
            END IF
            IF (DD==1) THEN
                IF( (sigma_1(1)>=sum(sigma_1(2:DD+1)*sigma_1(2:DD+1))) .and. (sigma_1(2)>=0) ) EXIT
            ELSEIF (DD==2) THEN
                IF( (sigma_1(1)>=sum(sigma_1(2:DD+1)*sigma_1(2:DD+1))) .and. (sigma_1(2)>=0) .and. (sigma_1(3)>=0) ) EXIT
            ELSEIF (DD==3) THEN
                IF( (sigma_1(1)>=sum(sigma_1(2:DD+1)*sigma_1(2:DD+1))) .and. (sigma_1(2)>=0) .and. (sigma_1(3)>=0)  &
                    &  .and. (sigma_1(4)>=0) ) EXIT
            ELSEIF (DD==4)  THEN
                IF( (sigma_1(1)>=sum(sigma_1(2:DD+1)*sigma_1(2:DD+1))) .and. (sigma_1(2)>=0) .and. (sigma_1(3)>=0) &
                    &  .and. (sigma_1(4)>=0)  .and. (sigma_1(5)>=0) ) EXIT
            END IF
        END DO
        DEALLOCATE(zero)


        Call OMP_SET_NUM_THREADS(Nthreads)
        !$omp parallel default(shared) private(g,NN, S1, V1, V2, Y, VV1, VV2, &
        !$omp &  zz_1, X, XX, W, ep_1, ep_2, ep_1v, ep_2v, IPIV, INFO, like_Y1, like_Y2, &
        !$omp & ithread, iblock, ibegin, iend)

        ithread = OMP_GET_THREAD_NUM()
        iblock = (GG+Nthreads-1)/Nthreads
        ibegin = ithread*iblock+1
        iend = min((ithread+1)*iblock,GG)

        DO g = ibegin,iend
            NN=N(g)
            ALLOCATE(Y(NN));
            ALLOCATE(X(NN,11)) ;
            ALLOCATE(XX(NN,22)) ;
            ALLOCATE(W(NN,NN));
            ALLOCATE(S1(NN,NN));
            ALLOCATE(V1(NN,NN));    ALLOCATE(V2(NN,NN))
            ALLOCATE(VV1(NN,NN));   ALLOCATE(VV2(NN,NN))
            ALLOCATE(zz_1(NN,DD))
            ALLOCATE(ep_1(NN));     ALLOCATE(ep_2(NN))
            ALLOCATE(ep_1v(NN));    ALLOCATE(ep_2v(NN))
            ALLOCATE(IPIV(NN))

            V1=(sigma_1(1)-sum(sigma_1(2:DD+1)*sigma_1(2:DD+1)))*EYE(NN)
            V2=(sigma_T(1,t-1)-sum(sigma_T(2:DD+1,t-1)*sigma_T(2:DD+1,t-1)))*EYE(NN)

            Y=group(1:NN,1,g)
            X=group(1:NN,2:12,g)
            W=network(1:NN,1:NN,g)
            XX(:,1:11)=X
            XX(:,12:22)=matmul(W,X)
            zz_1=zz(1:NN,1:DD,g)

            S1=eye(NN)-lambda_T(t)*W

            ep_1=matmul(S1,Y)-matmul(XX,beta_T(:,t))-matmul(zz_1,sigma_1(2:DD+1))-alpha_T(g,t-1)
            ep_2=matmul(S1,Y)-matmul(XX,beta_T(:,t))-matmul(zz_1,sigma_T(2:DD+1,t-1))-alpha_T(g,t-1)
            ep_1v=ep_1
            ep_2v=ep_2

            VV1=V1
            VV2=V2
            CALL SGESV(NN,1,VV1,NN,IPIV,ep_1v,NN,INFO)
            CALL SGESV(NN,1,VV2,NN,IPIV,ep_2v,NN,INFO)

            like_Y1=(-0.5)*NN*log(sigma_1(1)-sum(sigma_1(2:DD+1)*sigma_1(2:DD+1)))-dot_product(ep_1, ep_1v)/2
            like_Y2=(-0.5)*NN*log(sigma_T(1,t-1)-sum(sigma_T(2:DD+1,t-1)*sigma_T(2:DD+1,t-1)))-dot_product(ep_2, ep_2v)/2


            pp_sig(g)=(like_Y1-like_Y2)


            DEALLOCATE(Y);
            DEALLOCATE(X) ;
            DEALLOCATE(XX) ;
            DEALLOCATE(W);
            DEALLOCATE(S1);
            DEALLOCATE(V1); DEALLOCATE(V2)
            DEALLOCATE(VV1);  DEALLOCATE(VV2)
            DEALLOCATE(zz_1)
            DEALLOCATE(ep_1); DEALLOCATE(ep_2)
            DEALLOCATE(ep_1v); DEALLOCATE(ep_2v)
            DEALLOCATE(IPIV)
        END DO
        !$omp end parallel

        pp=sum(pp_sig)+logmvnpdf( sigma_1-sigma_0, Sig_0, 1+DD)-logmvnpdf(sigma_T(1:1+DD,t-1)-sigma_0, Sig_0, 1+DD)

        CALL random_number(rand)
        IF (log(rand)<=pp) THEN
            sigma_T(:,t)=sigma_1
            acc_4=acc_4+1.0
        ELSE
            sigma_T(:,t)=sigma_T(:,t-1)
        ENDIF
        acc_rate4(t)=acc_4/t
        !if ((acc_rate4(t)<0.05)) c_4=c_4/1.01
        !if ((acc_rate4(t)>0.15)) c_4=c_4*1.01

        !*************GIBBS STEP FOR ALPHA_G*********************!


        Call OMP_SET_NUM_THREADS(Nthreads)
        !$omp parallel default(shared) private(g,NN, Y, YY, &
        !$omp & X, XX, W, S1, zz_1, IPIV, R_g, &
        !$omp & ithread, iblock, ibegin, iend)

        ithread = OMP_GET_THREAD_NUM()
        iblock = (GG+Nthreads-1)/Nthreads
        ibegin = ithread*iblock+1
        iend = min((ithread+1)*iblock,GG)

        DO g = ibegin,iend

            NN=N(g)
            ALLOCATE(Y(NN));
            ALLOCATE(YY(NN));
            ALLOCATE(X(NN,11)) ;
            ALLOCATE(XX(NN,22)) ;
            ALLOCATE(W(NN,NN));
            ALLOCATE(S1(NN,NN));
            ALLOCATE(zz_1(NN,DD))

            R_g=(ALPHA_0**(-1)+(sigma_T(1,t)-sum(sigma_T(2:DD+1,t)*sigma_T(2:DD+1,t)))**(-1.0)*NN)**(-1.0)
            Y=group(1:NN,1,g)
            X=group(1:NN,2:12,g)
            W=network(1:NN,1:NN,g)
            XX(:,1:11)=X
            XX(:,12:22)=matmul(W,X)
            zz_1=zz(1:NN,1:DD,g)

            S1=eye(NN)-lambda_T(t)*W
            YY=matmul(S1,Y)-matmul(zz_1,sigma_T(2:DD+1,t))
            alpha_T(g,t)=R_g*(ALPHA_0**(-1)*a_0+(sigma_T(1,t)-sum(sigma_T(2:DD+1,t)*sigma_T(2:DD+1,t)))**(-1.0)*sum((YY-matmul(XX,beta_T(:,t)))))+random_normal()*sqrt(R_g)

            DEALLOCATE(Y);
            DEALLOCATE(YY)
            DEALLOCATE(X) ;
            DEALLOCATE(XX) ;
            DEALLOCATE(W);
            DEALLOCATE(S1);
            DEALLOCATE(zz_1);
        END DO
        !$omp end parallel


        !******** Calculate Log likelihood function value *********!

        Call OMP_SET_NUM_THREADS(Nthreads)
        !$omp parallel default(shared) private(g, NN,V, S, Y,X, XX, W,&
        !$omp & psi_1, ep_1, ep_1v, INFO, CODE, DD1,  &
        !$omp & C1, C2, C3, V1, zz_1, IPIV, INDX, det1, q_1,  &
        !$omp & pp, i, j, like_Y1, rand, &
        !$omp & ithread, iblock, ibegin, iend)

        ithread = OMP_GET_THREAD_NUM()
        iblock = (GG+Nthreads-1)/Nthreads
        ibegin = ithread*iblock+1
        iend = min((ithread+1)*iblock,GG)

        DO g = ibegin,iend

            NN=N(g)
            ALLOCATE(V(NN,NN));        ALLOCATE(S(NN,NN));
            ALLOCATE(Y(NN));
            ALLOCATE(X(NN,11)) ;       ALLOCATE(XX(NN,22)) ;
            ALLOCATE(W(NN,NN));
            ALLOCATE(V1(NN,NN));
            ALLOCATE(psi_1(NN,NN));
            ALLOCATE(ep_1(NN));
            ALLOCATE(ep_1v(NN));
            ALLOCATE(C1(NN,NN));       ALLOCATE(C2(NN,NN));
            ALLOCATE(C3(NN,NN));
            ALLOCATE(zz_1(NN,DD));
            ALLOCATE(IPIV(NN));        ALLOCATE(INDX(NN));

            V=(sigma_T(1,t)-sum(sigma_T(2:DD+1,t)*sigma_T(2:DD+1,t)))*EYE(NN)
            Y=group(1:NN,1,g)
            X=group(1:NN,2:12,g)
            W=network(1:NN,1:NN,g)
            XX(:,1:11)=X
            XX(:,12:22)=matmul(W,X)
            zz_1=zz(1:NN,1:DD,g)

            C1=grade(1:NN,1:NN,g)
            C2=sex(1:NN,1:NN,g)
            C3=race(1:NN,1:NN,g)

            pp=0.0
            DO i=1,NN
                DO j=1, NN
                    IF (j/=i) THEN
                        IF (DD==1) THEN
                            psi_1(i,j)=gamma_T(1,t)+gamma_T(2,t)*c1(i,j)+gamma_T(3,t)*c2(i,j)+gamma_T(4,t)*c3(i,j)-gamma_T(5,t)*abs(zz_1(i,1)-zz_1(j,1))
                        ELSEIF (DD==2) THEN
                            psi_1(i,j)=gamma_T(1,t)+gamma_T(2,t)*c1(i,j)+gamma_T(3,t)*c2(i,j)+gamma_T(4,t)*c3(i,j)-gamma_T(5,t)*abs(zz_1(i,1)-zz_1(j,1))-gamma_T(6,t)*abs(zz_1(i,2)-zz_1(j,2))
                        ELSEIF (DD==3) THEN
                            psi_1(i,j)=gamma_T(1,t)+gamma_T(2,t)*c1(i,j)+gamma_T(3,t)*c2(i,j)+gamma_T(4,t)*c3(i,j)-gamma_T(5,t)*abs(zz_1(i,1)-zz_1(j,1))-gamma_T(6,t)*abs(zz_1(i,2)-zz_1(j,2)) &
                                & -gamma_T(7,t)*abs(zz_1(i,3)-zz_1(j,3))
                        ELSEIF (DD==4) THEN
                            psi_1(i,j)=gamma_T(1,t)+gamma_T(2,t)*c1(i,j)+gamma_T(3,t)*c2(i,j)+gamma_T(4,t)*c3(i,j)-gamma_T(5,t)*abs(zz_1(i,1)-zz_1(j,1))-gamma_T(6,t)*abs(zz_1(i,2)-zz_1(j,2)) &
                                & -gamma_T(7,t)*abs(zz_1(i,3)-zz_1(j,3)) -gamma_T(8,t)*abs(zz_1(i,4)-zz_1(j,4))
                        ENDIF
                        q_1=psi_1(i,j)*W(i,j)-log(1+exp(psi_1(i,j)))
                        pp=pp+q_1
                    ENDIF
                END DO
            END DO

            S=eye(NN)-lambda_T(t)*W

            ep_1=matmul(S,Y)-matmul(XX,beta_T(:,t))-matmul(zz_1,sigma_T(2:DD+1,t))-alpha_T(g,t)
            ep_1v=ep_1

            V1=V
            CALL SGESV(NN,1,V1,NN,IPIV,ep_1v,NN,INFO)

            det1=FindDet(S, NN)

            like_Y1=-0.5*NN*log(2*pi)-0.5*NN*log(sigma_T(1,t)-sum(sigma_T(2:DD+1,t)*sigma_T(2:DD+1,t)))+log(det1)-dot_product(ep_1, ep_1v)/2

            pp_G(g)=pp+like_Y1

            DEALLOCATE(V);        DEALLOCATE(S);
            DEALLOCATE(Y);
            DEALLOCATE(X) ;       DEALLOCATE(XX) ;
            DEALLOCATE(W);
            DEALLOCATE(V1);
            DEALLOCATE(psi_1);
            DEALLOCATE(ep_1);
            DEALLOCATE(ep_1v);
            DEALLOCATE(C1);       DEALLOCATE(C2);
            DEALLOCATE(C3);
            DEALLOCATE(zz_1);
            DEALLOCATE(IPIV);     DEALLOCATE(INDX);
        END DO
        !$omp end parallel

        logp_T(t)=sum(pp_G)


        call system_clock ( clock_stop, clock_rate, clock_max )
        t_round=t/20
        IF (t-t_round*20==0) THEN
            WRITE(*,*)   'T=', t
            WRITE(*,'(a6, f7.4)') 'Time:', real ( clock_stop - clock_start ) / real ( clock_rate )
            WRITE(*,'(a6,12f7.4)') 'GAMMA:', gamma_T(:,t)
            WRITE(*,'(a6, f7.4)') 'LAMBDA:', lambda_T(t)
            WRITE(*,'(a6,6f7.4)') 'BETA:', beta_T(1:6,t)
            WRITE(*,'(a6,6f7.4)') 'BETA:', beta_T(7:11,t)
            WRITE(*,'(a6,5f7.4)') 'SIGMA:',  sigma_T(:,t)
            WRITE(*,'(a6,f10.2)') 'Likeli:',  logp_T(t)
            WRITE(*,'(a6,f7.4)') 'acc_1:',  acc_rate1(t)
            WRITE(*,'(a6,f7.4)') 'acc_2:',  acc_rate2(t)
            WRITE(*,'(a6,f7.4)') 'acc_4:',  acc_rate4(t)
            WRITE(*,'(a6,7f7.4)') 'acc_3:', acc_rate3(1,t),acc_rate3(11,t)
            WRITE(*,'(a6,7f7.4)') 'c_3:', c_3(1),c_3(11)
            WRITE(*,'(a6,7f7.4)') 'c_4:', c_4
            WRITE(*,'(a6,7f7.4)') 'zz:', zz(1,1:DD,1), zz(1,1:DD,2), zz(1,1:DD,3)
        END IF

        IF (DD==1) THEN
            WRITE(1000,'(5f12.4)')  gamma_T(:,t)
            WRITE(4000,'(2f12.4)')  sigma_T(:,t)
        ELSEIF (DD==2) THEN
            WRITE(1000,'(6f12.4)')  gamma_T(:,t)
            WRITE(4000,'(3f12.4)')  sigma_T(:,t)
        ELSEIF (DD==3) THEN
            WRITE(1000,'(7f12.4)')  gamma_T(:,t)
            WRITE(4000,'(4f12.4)')  sigma_T(:,t)
        ELSEIF (DD==4) THEN
            WRITE(1000,'(8f12.4)')  gamma_T(:,t)
            WRITE(4000,'(5f12.4)')  sigma_T(:,t)
        ENDIF

        WRITE(2000,'(1f10.4)')   lambda_T(t)
        WRITE(3000,'(22f10.4)')  beta_T(:,t)
        WRITE(6000,100)          alpha_T(:,t)
        WRITE(7000,'(f12.3)')    logp_T(t)

        !t_round=t/THIN
        !IF (t-t_round*THIN==0) THEN
        !    WRITE(5000,200)  ((zz(i,:,g),i=1,220),g=1,GG)
        !END IF


    END DO



    END PROGRAM estimation



