    INCLUDE 'mkl_vsl.f90'
    
    !**************************************************************************
    ! This code is written for "Smoking Initiation -- Peers and Personality",
    ! Chih-Sheng Hsieh & Hans van Kippersluis
    ! Quantitative Economics
    !
    ! Replication of Table A7: Peer Effects on Smoking Dummy and Frequency
    ! -- Row-normalized SC-SAR models
    !
    ! * users should change working directory in path1, path2, and path3
    ! * GG denotes the number of school networks
    ! * TT denotes the length of MCMC iteractions
    ! * MAX denotes the maximum size of school networks
    ! * DD denotes the dimension of latent variables
    ! * THIN denotes the thinning of the MCMC chain
    ! * Nthreads denotes the CPU threads for computing
    ! * H denotes the block size of updating latent variables
    ! * dep=2: smoking dummy
    ! * dep=3: smoking frequency
    ! * raw_norm=0: raw unnormalized weights matrix
    ! * raw_norm=1: raw normalized weights matrix
    ! * burn denotes the length of burn-in iterations
    ! * users should change initial gamma values w.r.t. DD in line 171-176
    !
    ! Please address any questions to:    
    ! Chih-Sheng Hsieh
    ! cshshieh@cuhk.edu.hk
    !***************************************************************************

    PROGRAM estimation

    USE random
    USE TOOL
    USE lapack95
    USE MKL_VSL_TYPE
    USE MKL_VSL
    USE omp_lib
    IMPLICIT NONE


    !****************** GENERAL VARIABLES ****************!
    CHARACTER(100) :: path1='D:/research/2014_health_study/'
    CHARACTER(100) :: path2='/data/inhome/smoke/neuroticism/'
    CHARACTER(100) :: path3='/result/SCSAR/smoke_frequency/homo/contextual_latent/dim4/rownorm'
    INTEGER, PARAMETER :: GG=118, TT=250000, MAX=200, DD=4, dep=3, THIN=1000, Nthreads=6, H=4, raw_norm=1, burn=100000
    REAL, PARAMETER    :: pi=3.14159265
    REAL               :: c_1, c_2, c_3, acc_1, acc_2, rand, acc_0v, acc_4, acc_3, c_0(GG), acc_0(GG), kk, AICM
    INTEGER            :: NN, N(GG),  k, censor(GG), flag(GG)
    INTEGER            :: brng, method, me, i, j, g, t, d, ier,  r, seedsize, seed(2), errorflag, INFO, seed_mkl, t_round
    INTEGER            :: clock_max, clock_rate, clock_start, clock_stop, ithread, iblock,ibegin,iend
    REAL, DIMENSION(MAX,MAX,GG) :: grade, sex, race, network, W_norm, W_FF, W_FM, W_MF, W_MM
    INTEGER, DIMENSION(MAX,GG) :: smoke_dummy
    REAL, DIMENSION(MAX,18,GG) :: group
    REAL, ALLOCATABLE :: C1(:,:), C2(:,:), C3(:,:), W(:,:), IPIV(:), Y(:), X(:,:)
    REAL, ALLOCATABLE :: Nu(:), Co(:), Ex(:), FACT(:,:)
    REAL                 :: DET1, DET2
    REAL(kind=8)         :: pp
    INTEGER, ALLOCATABLE :: INDX(:), p_dim(:), IPVT(:)
    INTEGER ::  errcode
    INTEGER :: CODE, DD1, DD2, SEED_PERM
    LOGICAL            :: first=.true.
    TYPE (VSL_STREAM_STATE) :: stream

    !*** hyperparameters in the prior distribution ***!
    REAL, DIMENSION(7+DD)           :: gamma_0
    REAL, DIMENSION(30+2*DD)        :: beta_0
    REAL, DIMENSION(7+DD,7+DD)      :: G_0
    REAL, DIMENSION(30+2*DD,30+2*DD):: B_0, INV_B_0
    REAL                            :: sigma_0
    REAL                            :: Sig_0
    REAL                            ::  ALPHA_0, a_0, eta_0
    INTEGER                         ::  rho_0

    !*** posterior draws of parameters *******!

    REAL, DIMENSION(7+DD,TT)       :: gamma_T
    REAL, DIMENSION(TT)            :: lambda_T
    REAL, DIMENSION(TT)            :: acc_rate1, acc_rate2, acc_rate3
    REAL(kind=8), DIMENSION(TT)    :: logp_T
    REAL, DIMENSION(30+2*DD, TT)   :: beta_T
    REAL, DIMENSION(TT)            :: sigma_T
    REAL, DIMENSION(GG,TT)         :: alpha_T
    REAL, DIMENSION(GG,TT)         :: acc_rate0

    !******  VARIABLE IN M-H FOR Z_G  *******************************!

    REAL, DIMENSION(MAX,DD,GG):: zz
    REAL, ALLOCATABLE  :: zz_1(:,:), zz_2(:,:), V(:,:), S1(:,:), S2(:,:), XX(:,:),  cov_zz(:,:), temp_zz(:)
    REAL, ALLOCATABLE  :: V1(:,:), V2(:,:)
    REAL, ALLOCATABLE  :: psi(:,:), psi_1(:,:), psi_2(:,:), zero(:)
    REAL, ALLOCATABLE  :: ep(:), ep_1(:), ep_2(:)
    REAL, ALLOCATABLE  :: ep_1v(:), ep_2v(:)
    REAL(kind=8) ::   q_1, q_2, like_Y1, like_Y2, pp_w, q

    !REAL, DIMENSION(MAX,DD,GG,TT/THIN):: zz_T

    !********VARIABLE IN M-H FOR GAMMA & LAMBDA   ***************!

    REAL      :: lambda_1, lambda_2, lambda_0, COV_lambda_1
    REAL      :: gamma_1(7+DD), gamma_2(7+DD), COV_gamma_1(7+DD,7+DD)
    REAL(kind=8)  :: pp_l(GG), pp_G(GG), pp_sig(GG)

    !*************VARIABLE IN M-H FOR BETA **************!

    REAL :: XVX(30+2*DD,30+2*DD,GG), XVY(30+2*DD,GG), INV_B(30+2*DD,30+2*DD), INV_B2(30+2*DD,30+2*DD), beta_temp(30+2*DD)
    REAL, ALLOCATABLE ::  XX2(:,:), YY(:), S(:,:)

    !*************VARIABLE IN M-H FOR SIGMA_E & SIGMA_EZ & ALPHA_G**************!

    REAL              ::  R_g, sigma_1
    REAL, ALLOCATABLE :: ep_v(:), VV1(:,:), VV2(:,:)
    INTEGER           :: rho_1, stack

    !*******************************************************************************************************!

    INTEGER :: N_1, N_2, T_1, T_2
    REAL, ALLOCATABLE:: rowsum_w_ff(:), rowsum_w_fm(:), rowsum_w_mf(:), rowsum_w_mm(:), rowsum_w_norm(:)

    !*******************************************************************************************************!

100 format (14f10.4)
200 format (4f10.4)

    seedsize=2
    seed(1)=2.1103e+005
    seed(2)=2.1103e+005
    CALL RANDOM_SEED
    CALL RANDOM_SEED(SIZE=seedsize)
    CALL RANDOM_SEED(PUT=seed)

    seed_mkl=12345
    brng=VSL_BRNG_MCG31
    method=VSL_RNG_METHOD_GAUSSIANMV_BOXMULLER
    me=VSL_MATRIX_STORAGE_FULL

    errcode=vslnewstream(stream,brng,seed_mkl)

    CALL networksize(GG, MAX, path1, path2, N)
    CALL define_variable(GG, grade, sex, race, network, group, N, path1, path2, MAX)

    W_norm=0.0
    W_FF=0.0
    W_FM=0.0
    W_MF=0.0
    W_MM=0.0
    zz=0.0
    lambda_0=0.0
    gamma_0=0.0
    beta_0=0.0
    sigma_0=0.0
    eta_0=0.1
    rho_0=2
    G_0=eye(7+DD)*100.0
    B_0=eye(30+2*DD)*10.0
    Sig_0=100.0
    CALL FINDInv(B_0,INV_B_0,30+2*DD,ier)
    ALPHA_0=400.0
    a_0=0.0
    censor=0
    smoke_dummy=0

    c_0=0.1
    c_3=0.1
    acc_1=0.0;acc_2=0.0;acc_3=0
    acc_rate1=0.0
    acc_rate2=0.0
    acc_rate3=0.0

    ! initial value of the draws !
    !gamma_T(:,1)=(/ -3.2356  ,  2.3110   , 0.3811  ,  0.8758 ,  -0.0375 ,  -0.0257 ,  -0.2598  , 10.1573/)   ! D=1 !
    !gamma_T(:,1)=(/   -1.5175  ,  2.4617 ,   0.3530  ,  0.9279 ,  -0.0608 ,  -0.0296  , -0.2535 ,   6.0569  ,  5.0701 /)  ! D=2 !
    !gamma_T(:,1)=(/  -0.1704   ,  2.4617 ,   0.3530  ,  0.9279 ,  -0.0608 ,  -0.0296  , -0.2535 ,   4.3, 3.8, 3.5/) !D=3!
    gamma_T(:,1)=(/    1.0183  ,  2.7424  ,  0.3513  ,  1.1234  , -0.0522  , -0.0401 ,  -0.2608  ,  3.3805   , 3.2391  ,  3.0714  ,  2.8158/) !D=4!
    !gamma_T(:,1)=(/  2.0994  ,  2.8267  ,  0.3637  ,  1.2037  , -0.0641  , -0.0543 ,  -0.2749  ,  3.2886  ,  3.1909  ,  2.948 ,   2.8210  ,  2.3497  /) !D=5!
    !gamma_T(:,1)=(/ 2.9985 ,   2.8869  ,  0.3653  ,  1.2320 ,  -0.0615 ,  -0.0441 ,  -0.2823 ,   3.4217  ,  2.6503  ,  2.3376  ,  2.1337 ,   1.9016  ,  1.6863 /) !D=6!
    sigma_T(1)=0.2


    OPEN (1000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'gamma.txt', status='unknown')
    OPEN (2000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'lambda.txt', status='unknown')
    OPEN (3000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'beta.txt', status='unknown')
    OPEN (4000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'sigma.txt', status='unknown')
    !OPEN (5000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'latent_z.txt', status='replace')
    !OPEN (6000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'alpha.txt', status='unknown')
    OPEN (7000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) // 'loglikeli.txt', status='unknown')

    ! read raw data and sort by gender -> four spatial weight matrix !

    DO g=1,GG
        N_1=0
        N_2=0
        NN=N(g)
        ALLOCATE(W(NN,NN));
        ALLOCATE(p_dim(NN));
        ALLOCATE(rowsum_w_ff(NN))
        ALLOCATE(rowsum_w_fm(NN))
        ALLOCATE(rowsum_w_mf(NN))
        ALLOCATE(rowsum_w_mm(NN))
        ALLOCATE(rowsum_w_norm(NN))

        p_dim=group(1:NN,1,g)
        W=network(1:NN,1:NN,g)

        DO i=1,NN
            IF (p_dim(i)==0) N_1=N_1+1
            IF (p_dim(i)==1) N_2=N_2+1
        END DO
        T_1=N_1
        T_2=N_1+N_2

        w_ff(1:T_1,1:T_1,g)=W(1:T_1,1:T_1)
        w_fm(1:T_1,T_1+1:T_2,g)=W(1:T_1,T_1+1:T_2)
        w_mf(T_1+1:T_2,1:T_1,g)=W(T_1+1:T_2,1:T_1)
        w_mm(T_1+1:T_2,T_1+1:T_2,g)=W(T_1+1:T_2,T_1+1:T_2)
        w_norm(1:NN,1:NN,g)=W

        if (raw_norm==1) THEN
            rowsum_w_ff=sum(w_ff(1:NN,1:NN,g), dim=2)
            rowsum_w_fm=sum(w_fm(1:NN,1:NN,g), dim=2)
            rowsum_w_mf=sum(w_mf(1:NN,1:NN,g), dim=2)
            rowsum_w_mm=sum(w_mm(1:NN,1:NN,g), dim=2)
            rowsum_w_norm=sum(W, dim=2)
            DO i=1,NN
                if (rowsum_w_ff(i)==0) rowsum_w_ff(i)=1;
                if (rowsum_w_fm(i)==0) rowsum_w_fm(i)=1;
                if (rowsum_w_mf(i)==0) rowsum_w_mf(i)=1;
                if (rowsum_w_mm(i)==0) rowsum_w_mm(i)=1;
                if (rowsum_w_norm(i)==0) rowsum_w_norm(i)=1;

                w_ff(i,1:NN,g)=w_ff(i,1:NN,g)/rowsum_w_ff(i);
                w_fm(i,1:NN,g)=w_fm(i,1:NN,g)/rowsum_w_fm(i);
                w_mf(i,1:NN,g)=w_mf(i,1:NN,g)/rowsum_w_mf(i);
                w_mm(i,1:NN,g)=w_mm(i,1:NN,g)/rowsum_w_mm(i);
                w_norm(i,1:NN,g)=W(i,1:NN)/rowsum_w_norm(i);
            END DO
        END IF

        DO i=1,NN
            IF (group(i,dep,g)>=0.625) smoke_dummy(i,g)=1
        END DO

        DEALLOCATE(W,p_dim,rowsum_w_ff,rowsum_w_fm,rowsum_w_mf,rowsum_w_mm,rowsum_w_norm)
    END DO




    DO t=2, TT

        call system_clock ( clock_start, clock_rate, clock_max )

        DO
            lambda_1=random_normal()*0.1+lambda_T(t-1)
            IF (t>2) THEN
                lambda_2=random_normal()*sqrt(var(lambda_T(1:t-1),t-1))*2.38+lambda_T(t-1)
                lambda_1=0.95*lambda_2+0.05*lambda_1
            END IF
            IF ((lambda_1>=-1.0) .and. (lambda_1<=1.0)) EXIT
        END DO

        DO
            COV_gamma_1=EYE(7+DD)*0.0003/(7+DD)
            call spotrf('U',7+DD, COV_gamma_1, 7+DD, info)
            errcode=vsrnggaussianmv(method,stream,1,gamma_1, 7+DD, me, gamma_0, COV_gamma_1)
            gamma_1=gamma_1+gamma_T(:,t-1)
            IF (t>1000) THEN
                COV_gamma_1=cov(gamma_T(:,1:t-1),7+DD,t-1)*2.38**2/(7+DD)
                call spotrf('U',7+DD, COV_gamma_1, 7+DD, info)
                errcode=vsrnggaussianmv(method,stream,1,gamma_2, 7+DD, me, gamma_0, COV_gamma_1)
                gamma_2=gamma_2+gamma_T(:,t-1)
                gamma_1=0.95*gamma_2+0.05*gamma_1
            END IF
            IF (DD==1) THEN
                IF (gamma_1(7+DD)>=0) EXIT
            ELSEIF (DD==2) THEN
                IF (gamma_1(6+DD)>=gamma_1(7+DD)) EXIT
            ELSEIF (DD==3) THEN
                IF ((gamma_1(5+DD)>=gamma_1(6+DD)) .and. (gamma_1(6+DD)>=gamma_1(7+DD)) .and. (gamma_1(7+DD)>=0)) EXIT
            ELSEIF (DD==4)  THEN
                IF ((gamma_1(4+DD)>=gamma_1(5+DD)) .and. (gamma_1(5+DD)>=gamma_1(6+DD)) .and. (gamma_1(6+DD)>=gamma_1(7+DD)) .and. (gamma_1(7+DD)>=0)) EXIT
            ELSEIF (DD==5)  THEN
                IF ((gamma_1(3+DD)>=gamma_1(4+DD)) .and. (gamma_1(4+DD)>=gamma_1(5+DD)) .and. (gamma_1(5+DD)>=gamma_1(6+DD)) .and. (gamma_1(6+DD)>=gamma_1(7+DD)) .and. (gamma_1(7+DD)>=0)) EXIT
            ELSEIF (DD==6)  THEN
                IF ((gamma_1(2+DD)>=gamma_1(3+DD)) .and. (gamma_1(3+DD)>=gamma_1(4+DD)) .and. (gamma_1(4+DD)>=gamma_1(5+DD)) .and. (gamma_1(5+DD)>=gamma_1(6+DD)) .and. (gamma_1(6+DD)>=gamma_1(7+DD)) .and. (gamma_1(7+DD)>=0)) EXIT
            END IF
        END DO

        gamma_2=gamma_T(:,t-1)

        ! THE M-H ALGORITHM FOR SAMPLING Z !


        Call OMP_SET_NUM_THREADS(Nthreads)
        !$omp parallel default(shared) private(g, NN,V, S, Y, Nu, Co, Ex, X, XX, W,  IPVT, FACT, &
        !$omp & psi_1, psi_2, ep, ep_1, ep_2, ep_1v, ep_2v, S1, S2, INFO, CODE, DD1, DD2, &
        !$omp & C1, C2, C3, V1, V2, zz_1, zz_2, IPIV, INDX, DET1, DET2, q_1, q_2, &
        !$omp & kk, acc_0v, k, cov_zz, temp_zz, zero, d, pp, i, j, like_Y1, like_Y2, rand, &
        !$omp & ithread, iblock, ibegin, iend)

        ithread = OMP_GET_THREAD_NUM()
        iblock = (GG+Nthreads-1)/Nthreads
        ibegin = ithread*iblock+1
        iend = min((ithread+1)*iblock,GG)

        DO g = ibegin,iend

            NN=N(g)
            ALLOCATE(V(NN,NN));        ALLOCATE(S(NN,NN));
            ALLOCATE(S1(NN,NN));       ALLOCATE(S2(NN,NN));
            ALLOCATE(Nu(NN));
            ALLOCATE(Co(NN));          ALLOCATE(Ex(NN));
            ALLOCATE(Y(NN));
            ALLOCATE(X(NN,15)) ;       ALLOCATE(XX(NN,30+2*DD)) ;
            ALLOCATE(W(NN,NN));
            ALLOCATE(psi_1(NN,NN));    ALLOCATE(psi_2(NN,NN))
            ALLOCATE(ep(NN));
            ALLOCATE(ep_1(NN));        ALLOCATE(ep_2(NN));
            ALLOCATE(ep_1v(NN));       ALLOCATE(ep_2v(NN))  ;
            ALLOCATE(C1(NN,NN));       ALLOCATE(C2(NN,NN));
            ALLOCATE(C3(NN,NN));
            ALLOCATE(V1(NN,NN));       ALLOCATE(V2(NN,NN))  ;
            ALLOCATE(zz_1(NN,DD));     ALLOCATE(zz_2(NN,DD))
            ALLOCATE(IPIV(NN));        ALLOCATE(INDX(NN));
            ALLOCATE(FACT(NN,NN));     ALLOCATE(IPVT(NN))

            V=sigma_T(t-1)*EYE(NN)
            Y=group(1:NN,dep,g)
            Nu=group(1:NN,4,g)
            Co=group(1:NN,5,g)
            Ex=group(1:NN,6,g)
            X=group(1:NN,4:18,g)
            W=network(1:NN,1:NN,g)
            XX(:,1:15)=X
            XX(:,16:30)=matmul(w_norm(1:NN,1:NN,g),X)
            zz_1=zz(1:NN,1:DD,g)
            zz_2=zz(1:NN,1:DD,g)

            C1=grade(1:NN,1:NN,g)
            C2=sex(1:NN,1:NN,g)
            C3=race(1:NN,1:NN,g)

            S=eye(NN)-lambda_T(t-1)*w_norm(1:NN,1:NN,g)

            kk=NN
            kk=floor(kk/H)
            IF (kk==0) kk=kk+1

            acc_0v=0.0
            DO k=1, kk
                IF(k==kk) THEN
                    ALLOCATE(cov_zz(NN-(k-1)*H,NN-(k-1)*H))
                    ALLOCATE(temp_zz(NN-(k-1)*H))
                    ALLOCATE(zero(NN-(k-1)*H))
                    zero=0.0
                    cov_zz=EYE(NN-(k-1)*H)*c_0(g)
                    call spotrf('U',NN-(k-1)*H, cov_zz, NN-(k-1)*H, info)
                    DO d=1, DD
                        errcode=vsrnggaussianmv(method,stream, 1, temp_zz, NN-(k-1)*H, me, zero, cov_zz)
                        zz_1((k-1)*H+1:NN,d)=temp_zz+zz_2((k-1)*H+1:NN,d)
                    END DO
                    DEALLOCATE(cov_zz)
                    DEALLOCATE(temp_zz)
                    DEALLOCATE(zero)
                ELSE
                    ALLOCATE(cov_zz(H,H))
                    ALLOCATE(temp_zz(H))
                    ALLOCATE(zero(H))
                    zero=0.0
                    cov_zz=EYE(H)*c_0(g)
                    call spotrf('U',H, cov_zz, H, info)
                    DO d=1,DD
                        errcode=vsrnggaussianmv(method,stream, 1, temp_zz, H, me, zero, cov_zz)
                        zz_1((k-1)*H+1:k*H,d)=temp_zz+zz_2((k-1)*H+1:k*H,d)
                    END DO
                    DEALLOCATE(cov_zz)
                    DEALLOCATE(temp_zz)
                    DEALLOCATE(zero)
                END IF

                pp=0.0
                DO i=1,NN
                    DO j=1,NN
                        IF (j/=i) THEN
                            IF (DD==1) THEN
                                psi_1(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j) &
                                    & +gamma_T(5,t-1)*abs(Nu(i)-Nu(j))+gamma_T(6,t-1)*abs(Co(i)-Co(j))+gamma_T(7,t-1)*abs(Ex(i)-Ex(j)) &
                                    & -gamma_T(8,t-1)*abs(zz_1(i,1)-zz_1(j,1))
                                psi_2(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j) &
                                    & +gamma_T(5,t-1)*abs(Nu(i)-Nu(j))+gamma_T(6,t-1)*abs(Co(i)-Co(j))+gamma_T(7,t-1)*abs(Ex(i)-Ex(j)) &
                                    & -gamma_T(8,t-1)*abs(zz_2(i,1) -zz_2(j,1))
                            ELSEIF (DD==2) THEN
                                psi_1(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j) &
                                    & +gamma_T(5,t-1)*abs(Nu(i)-Nu(j))+gamma_T(6,t-1)*abs(Co(i)-Co(j))+gamma_T(7,t-1)*abs(Ex(i)-Ex(j)) &
                                    & -gamma_T(8,t-1)*abs(zz_1(i,1)-zz_1(j,1))-gamma_T(9,t-1)*abs(zz_1(i,2)-zz_1(j,2))
                                psi_2(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j) &
                                    & +gamma_T(5,t-1)*abs(Nu(i)-Nu(j))+gamma_T(6,t-1)*abs(Co(i)-Co(j))+gamma_T(7,t-1)*abs(Ex(i)-Ex(j)) &
                                    & -gamma_T(8,t-1)*abs(zz_2(i,1)-zz_2(j,1))-gamma_T(9,t-1)*abs(zz_2(i,2)-zz_2(j,2))
                            ELSEIF (DD==3) THEN
                                psi_1(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j)  &
                                    & +gamma_T(5,t-1)*abs(Nu(i)-Nu(j))+gamma_T(6,t-1)*abs(Co(i)-Co(j))+gamma_T(7,t-1)*abs(Ex(i)-Ex(j)) &
                                    & -gamma_T(8,t-1)*abs(zz_1(i,1)-zz_1(j,1)) -gamma_T(9,t-1)*abs(zz_1(i,2)-zz_1(j,2)) &
                                    & -gamma_T(10,t-1)*abs(zz_1(i,3)-zz_1(j,3))
                                psi_2(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j)  &
                                    & +gamma_T(5,t-1)*abs(Nu(i)-Nu(j))+gamma_T(6,t-1)*abs(Co(i)-Co(j))+gamma_T(7,t-1)*abs(Ex(i)-Ex(j)) &
                                    & -gamma_T(8,t-1)*abs(zz_2(i,1)-zz_2(j,1)) -gamma_T(9,t-1)*abs(zz_2(i,2)-zz_2(j,2)) &
                                    & -gamma_T(10,t-1)*abs(zz_2(i,3)-zz_2(j,3))
                            ELSEIF (DD==4) THEN
                                psi_1(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j)  &
                                    & +gamma_T(5,t-1)*abs(Nu(i)-Nu(j))+gamma_T(6,t-1)*abs(Co(i)-Co(j))+gamma_T(7,t-1)*abs(Ex(i)-Ex(j)) &
                                    & -gamma_T(8,t-1)*abs(zz_1(i,1)-zz_1(j,1))-gamma_T(9,t-1)*abs(zz_1(i,2)-zz_1(j,2)) &
                                    & -gamma_T(10,t-1)*abs(zz_1(i,3)-zz_1(j,3))-gamma_T(11,t-1)*abs(zz_1(i,4)-zz_1(j,4))
                                psi_2(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j)  &
                                    & +gamma_T(5,t-1)*abs(Nu(i)-Nu(j))+gamma_T(6,t-1)*abs(Co(i)-Co(j))+gamma_T(7,t-1)*abs(Ex(i)-Ex(j)) &
                                    & -gamma_T(8,t-1)*abs(zz_2(i,1)-zz_2(j,1)) -gamma_T(9,t-1)*abs(zz_2(i,2)-zz_2(j,2))  &
                                    & -gamma_T(10,t-1)*abs(zz_2(i,3)-zz_2(j,3)) -gamma_T(11,t-1)*abs(zz_2(i,4)-zz_2(j,4))
                            ELSEIF (DD==5) THEN
                                psi_1(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j)  &
                                    & +gamma_T(5,t-1)*abs(Nu(i)-Nu(j))+gamma_T(6,t-1)*abs(Co(i)-Co(j))+gamma_T(7,t-1)*abs(Ex(i)-Ex(j)) &
                                    & -gamma_T(8,t-1)*abs(zz_1(i,1)-zz_1(j,1))-gamma_T(9,t-1)*abs(zz_1(i,2)-zz_1(j,2)) &
                                    & -gamma_T(10,t-1)*abs(zz_1(i,3)-zz_1(j,3))-gamma_T(11,t-1)*abs(zz_1(i,4)-zz_1(j,4)) &
                                    & -gamma_T(12,t-1)*abs(zz_1(i,5)-zz_1(j,5))
                                psi_2(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j)  &
                                    & +gamma_T(5,t-1)*abs(Nu(i)-Nu(j))+gamma_T(6,t-1)*abs(Co(i)-Co(j))+gamma_T(7,t-1)*abs(Ex(i)-Ex(j)) &
                                    & -gamma_T(8,t-1)*abs(zz_2(i,1)-zz_2(j,1)) -gamma_T(9,t-1)*abs(zz_2(i,2)-zz_2(j,2))  &
                                    & -gamma_T(10,t-1)*abs(zz_2(i,3)-zz_2(j,3)) -gamma_T(11,t-1)*abs(zz_2(i,4)-zz_2(j,4)) &
                                    & -gamma_T(12,t-1)*abs(zz_2(i,5)-zz_2(j,5))
                            ELSEIF (DD==6) THEN
                                psi_1(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j)  &
                                    & +gamma_T(5,t-1)*abs(Nu(i)-Nu(j))+gamma_T(6,t-1)*abs(Co(i)-Co(j))+gamma_T(7,t-1)*abs(Ex(i)-Ex(j)) &
                                    & -gamma_T(8,t-1)*abs(zz_1(i,1)-zz_1(j,1))-gamma_T(9,t-1)*abs(zz_1(i,2)-zz_1(j,2)) &
                                    & -gamma_T(10,t-1)*abs(zz_1(i,3)-zz_1(j,3))-gamma_T(11,t-1)*abs(zz_1(i,4)-zz_1(j,4)) &
                                    & -gamma_T(12,t-1)*abs(zz_1(i,5)-zz_1(j,5))-gamma_T(13,t-1)*abs(zz_1(i,6)-zz_1(j,6))
                                psi_2(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C1(i,j)+gamma_T(3,t-1)*C2(i,j)+gamma_T(4,t-1)*C3(i,j)  &
                                    & +gamma_T(5,t-1)*abs(Nu(i)-Nu(j))+gamma_T(6,t-1)*abs(Co(i)-Co(j))+gamma_T(7,t-1)*abs(Ex(i)-Ex(j)) &
                                    & -gamma_T(8,t-1)*abs(zz_2(i,1)-zz_2(j,1)) -gamma_T(9,t-1)*abs(zz_2(i,2)-zz_2(j,2))  &
                                    & -gamma_T(10,t-1)*abs(zz_2(i,3)-zz_2(j,3)) -gamma_T(11,t-1)*abs(zz_2(i,4)-zz_2(j,4)) &
                                    & -gamma_T(12,t-1)*abs(zz_2(i,5)-zz_2(j,5))-gamma_T(13,t-1)*abs(zz_2(i,6)-zz_2(j,6))
                            ENDIF

                            q_1=psi_1(i,j)*W(i,j)-log(1+exp(psi_1(i,j)))
                            q_2=psi_2(i,j)*W(i,j)-log(1+exp(psi_2(i,j)))
                            pp=pp+(q_1-q_2)
                        ENDIF
                    END DO
                END DO

                XX(:,31:30+DD)=zz_1
                XX(:,31+DD:30+2*DD)=matmul(w_norm(1:NN,1:NN,g),zz_1)
                ep_1=matmul(S,Y)-matmul(XX,beta_T(:,t-1))-alpha_T(g,t-1)

                XX(:,31:30+DD)=zz_2
                XX(:,31+DD:30+2*DD)=matmul(w_norm(1:NN,1:NN,g),zz_2)
                ep_2=matmul(S,Y)-matmul(XX,beta_T(:,t-1))-alpha_T(g,t-1)


                like_Y1=-dot_product(ep_1, ep_1)/(2*sigma_T(t-1))
                like_Y2=-dot_product(ep_2, ep_2)/(2*sigma_T(t-1))

                IF (k==kk) THEN
                    pp=pp+(like_Y1-like_Y2)+logmvnpdf(reshape(zz_1((k-1)*H+1:NN,:),(/(NN-(k-1)*H)*DD,1/)),eye((NN-(k-1)*H)*DD),(NN-(k-1)*H)*DD)-logmvnpdf(reshape(zz_2((k-1)*H+1:NN,:),(/(NN-(k-1)*H)*DD,1/)),eye((NN-(k-1)*H)*DD),(NN-(k-1)*H)*DD)
                ELSE
                    pp=pp+(like_Y1-like_Y2)+logmvnpdf(reshape(zz_1((k-1)*H+1:k*H,:),(/H*DD,1/)),eye(H*DD),H*DD)-logmvnpdf(reshape(zz_2((k-1)*H+1:k*H,:),(/H*DD,1/)),eye(H*DD),H*DD)
                END IF

                CALL random_number(rand)
                IF (log(rand)<=pp) THEN
                    zz_2=zz_1
                    acc_0v=acc_0v+1
                END IF
                zz_1=zz_2
            END DO
            if (3*acc_0v>2*kk) acc_0(g)=acc_0(g)+1
            acc_rate0(g,t)=acc_0(g)/t
            if ((acc_rate0(g,t)<0.2) )  c_0(g)=c_0(g)/1.01
            if ((acc_rate0(g,t)>0.3) )  c_0(g)=c_0(g)*1.01
            zz(1:NN,1:DD,g)=zz_1

            !***** THE M-H ALGORITHM FOR SAMPLING GAMMA & LAMBDA ******!

            pp=0.0
            DO i=1,NN
                DO j=1, NN
                    IF (j/=i) THEN
                        IF (DD==1) THEN
                            psi_1(i,j)=gamma_1(1)+gamma_1(2)*c1(i,j)+gamma_1(3)*c2(i,j)+gamma_1(4)*c3(i,j) &
                                & +gamma_1(5)*abs(Nu(i)-Nu(j))+gamma_1(6)*abs(Co(i)-Co(j))+gamma_1(7)*abs(Ex(i)-Ex(j))-gamma_1(8)*abs(zz_1(i,1)-zz_1(j,1))
                            psi_2(i,j)=gamma_2(1)+gamma_2(2)*c1(i,j)+gamma_2(3)*c2(i,j)+gamma_2(4)*c3(i,j) &
                                & +gamma_2(5)*abs(Nu(i)-Nu(j))+gamma_2(6)*abs(Co(i)-Co(j))+gamma_2(7)*abs(Ex(i)-Ex(j))-gamma_2(8)*abs(zz_1(i,1)-zz_1(j,1))
                        ELSEIF (DD==2) THEN
                            psi_1(i,j)=gamma_1(1)+gamma_1(2)*c1(i,j)+gamma_1(3)*c2(i,j)+gamma_1(4)*c3(i,j) &
                                & +gamma_1(5)*abs(Nu(i)-Nu(j))+gamma_1(6)*abs(Co(i)-Co(j))+gamma_1(7)*abs(Ex(i)-Ex(j))-gamma_1(8)*abs(zz_1(i,1)-zz_1(j,1)) &
                                & -gamma_1(9)*abs(zz_1(i,2)-zz_1(j,2))
                            psi_2(i,j)=gamma_2(1)+gamma_2(2)*c1(i,j)+gamma_2(3)*c2(i,j)+gamma_2(4)*c3(i,j) &
                                & +gamma_2(5)*abs(Nu(i)-Nu(j))+gamma_2(6)*abs(Co(i)-Co(j))+gamma_2(7)*abs(Ex(i)-Ex(j))-gamma_2(8)*abs(zz_1(i,1)-zz_1(j,1)) &
                                & -gamma_2(9)*abs(zz_1(i,2)-zz_1(j,2))
                        ELSEIF (DD==3) THEN
                            psi_1(i,j)=gamma_1(1)+gamma_1(2)*c1(i,j)+gamma_1(3)*c2(i,j)+gamma_1(4)*c3(i,j) &
                                & +gamma_1(5)*abs(Nu(i)-Nu(j))+gamma_1(6)*abs(Co(i)-Co(j))+gamma_1(7)*abs(Ex(i)-Ex(j))-gamma_1(8)*abs(zz_1(i,1)-zz_1(j,1))-gamma_1(9)*abs(zz_1(i,2)-zz_1(j,2)) &
                                &-gamma_1(10)*abs(zz_1(i,3)-zz_1(j,3))
                            psi_2(i,j)=gamma_2(1)+gamma_2(2)*c1(i,j)+gamma_2(3)*c2(i,j)+gamma_2(4)*c3(i,j) &
                                & +gamma_2(5)*abs(Nu(i)-Nu(j))+gamma_2(6)*abs(Co(i)-Co(j))+gamma_2(7)*abs(Ex(i)-Ex(j))-gamma_2(8)*abs(zz_1(i,1)-zz_1(j,1))-gamma_2(9)*abs(zz_1(i,2)-zz_1(j,2)) &
                                &-gamma_2(10)*abs(zz_1(i,3)-zz_1(j,3))
                        ELSEIF (DD==4) THEN
                            psi_1(i,j)=gamma_1(1)+gamma_1(2)*c1(i,j)+gamma_1(3)*c2(i,j)+gamma_1(4)*c3(i,j)&
                                & +gamma_1(5)*abs(Nu(i)-Nu(j))+gamma_1(6)*abs(Co(i)-Co(j))+gamma_1(7)*abs(Ex(i)-Ex(j))-gamma_1(8)*abs(zz_1(i,1)-zz_1(j,1))-gamma_1(9)*abs(zz_1(i,2)-zz_1(j,2)) &
                                & -gamma_1(10)*abs(zz_1(i,3)-zz_1(j,3))-gamma_1(11)*abs(zz_1(i,4)-zz_1(j,4))
                            psi_2(i,j)=gamma_2(1)+gamma_2(2)*c1(i,j)+gamma_2(3)*c2(i,j)+gamma_2(4)*c3(i,j)&
                                & +gamma_2(5)*abs(Nu(i)-Nu(j))+gamma_2(6)*abs(Co(i)-Co(j))+gamma_2(7)*abs(Ex(i)-Ex(j))-gamma_2(8)*abs(zz_1(i,1)-zz_1(j,1))-gamma_2(9)*abs(zz_1(i,2)-zz_1(j,2)) &
                                & -gamma_2(10)*abs(zz_1(i,3)-zz_1(j,3))-gamma_2(11)*abs(zz_1(i,4)-zz_1(j,4))
                        ELSEIF (DD==5) THEN
                            psi_1(i,j)=gamma_1(1)+gamma_1(2)*c1(i,j)+gamma_1(3)*c2(i,j)+gamma_1(4)*c3(i,j)&
                                & +gamma_1(5)*abs(Nu(i)-Nu(j))+gamma_1(6)*abs(Co(i)-Co(j))+gamma_1(7)*abs(Ex(i)-Ex(j))-gamma_1(8)*abs(zz_1(i,1)-zz_1(j,1))-gamma_1(9)*abs(zz_1(i,2)-zz_1(j,2)) &
                                & -gamma_1(10)*abs(zz_1(i,3)-zz_1(j,3))-gamma_1(11)*abs(zz_1(i,4)-zz_1(j,4))-gamma_1(12)*abs(zz_1(i,5)-zz_1(j,5))
                            psi_2(i,j)=gamma_2(1)+gamma_2(2)*c1(i,j)+gamma_2(3)*c2(i,j)+gamma_2(4)*c3(i,j)&
                                & +gamma_2(5)*abs(Nu(i)-Nu(j))+gamma_2(6)*abs(Co(i)-Co(j))+gamma_2(7)*abs(Ex(i)-Ex(j))-gamma_2(8)*abs(zz_1(i,1)-zz_1(j,1))-gamma_2(9)*abs(zz_1(i,2)-zz_1(j,2)) &
                                & -gamma_2(10)*abs(zz_1(i,3)-zz_1(j,3))-gamma_2(11)*abs(zz_1(i,4)-zz_1(j,4))-gamma_2(12)*abs(zz_1(i,5)-zz_1(j,5))
                        ELSEIF (DD==6) THEN
                            psi_1(i,j)=gamma_1(1)+gamma_1(2)*c1(i,j)+gamma_1(3)*c2(i,j)+gamma_1(4)*c3(i,j)&
                                & +gamma_1(5)*abs(Nu(i)-Nu(j))+gamma_1(6)*abs(Co(i)-Co(j))+gamma_1(7)*abs(Ex(i)-Ex(j))-gamma_1(8)*abs(zz_1(i,1)-zz_1(j,1))-gamma_1(9)*abs(zz_1(i,2)-zz_1(j,2)) &
                                & -gamma_1(10)*abs(zz_1(i,3)-zz_1(j,3))-gamma_1(11)*abs(zz_1(i,4)-zz_1(j,4))-gamma_1(12)*abs(zz_1(i,5)-zz_1(j,5)) -gamma_1(13)*abs(zz_1(i,6)-zz_1(j,6))
                            psi_2(i,j)=gamma_2(1)+gamma_2(2)*c1(i,j)+gamma_2(3)*c2(i,j)+gamma_2(4)*c3(i,j)&
                                & +gamma_2(5)*abs(Nu(i)-Nu(j))+gamma_2(6)*abs(Co(i)-Co(j))+gamma_2(7)*abs(Ex(i)-Ex(j))-gamma_2(8)*abs(zz_1(i,1)-zz_1(j,1))-gamma_2(9)*abs(zz_1(i,2)-zz_1(j,2)) &
                                & -gamma_2(10)*abs(zz_1(i,3)-zz_1(j,3))-gamma_2(11)*abs(zz_1(i,4)-zz_1(j,4))-gamma_2(12)*abs(zz_1(i,5)-zz_1(j,5)) -gamma_2(13)*abs(zz_1(i,6)-zz_1(j,6))
                        ENDIF
                        q_1=psi_1(i,j)*W(i,j)-log(1+exp(psi_1(i,j)))
                        q_2=psi_2(i,j)*W(i,j)-log(1+exp(psi_2(i,j)))
                        pp=pp+(q_1-q_2)
                    ENDIF
                END DO
            END DO

            pp_G(g)=pp


            S1=eye(NN)-lambda_1*w_norm(1:NN,1:NN,g)
            S2=eye(NN)-lambda_T(t-1)*w_norm(1:NN,1:NN,g)

            XX(:,31:30+DD)=zz_1
            XX(:,31+DD:30+2*DD)=matmul(w_norm(1:NN,1:NN,g),zz_1)
            ep_1=matmul(S1,Y)-matmul(XX,beta_T(:,t-1))-alpha_T(g,t-1)
            ep_2=matmul(S2,Y)-matmul(XX,beta_T(:,t-1))-alpha_T(g,t-1)

            like_Y1=log(FindDet(S1,NN))-dot_product(ep_1, ep_1)/(2*sigma_T(t-1))
            like_Y2=log(FindDet(S2,NN))-dot_product(ep_2, ep_2)/(2*sigma_T(t-1))
            pp_l(g)=(like_Y1-like_Y2)

            DEALLOCATE(V);        DEALLOCATE(S);
            DEALLOCATE(S1);       DEALLOCATE(S2);
            DEALLOCATE(Y);
            DEALLOCATE(Nu);
            DEALLOCATE(Co);
            DEALLOCATE(Ex);
            DEALLOCATE(X) ;       DEALLOCATE(XX) ;
            DEALLOCATE(W);
            DEALLOCATE(psi_1);    DEALLOCATE(psi_2)
            DEALLOCATE(ep);
            DEALLOCATE(ep_1);     DEALLOCATE(ep_2);
            DEALLOCATE(ep_1v);    DEALLOCATE(ep_2v)  ;
            DEALLOCATE(C1);       DEALLOCATE(C2);
            DEALLOCATE(C3);
            DEALLOCATE(V1);       DEALLOCATE(V2)  ;
            DEALLOCATE(zz_1);     DEALLOCATE(zz_2)
            DEALLOCATE(IPIV);     DEALLOCATE(INDX);
            DEALLOCATE(FACT);     DEALLOCATE(IPVT)

        END DO
        !$omp end parallel

        pp=sum(pp_G)+logmvnpdf(gamma_1, G_0, 7+DD)-logmvnpdf(gamma_2, G_0, 7+DD)

        CALL random_number(rand)
        IF (log(rand)<=pp) THEN
            gamma_T(:,t)=gamma_1
            acc_1=acc_1+1.0
        ELSE
            gamma_T(:,t)=gamma_T(:,t-1)
        ENDIF
        acc_rate1(t)=acc_1/t

        CALL random_number(rand)
        IF (log(rand)<=sum(pp_l)) THEN
            lambda_T(t)=lambda_1
            acc_2=acc_2+1.0
        ELSE
            lambda_T(t)=lambda_T(t-1)
        END IF
        acc_rate2(t)=acc_2/t



        !***************GIBBS STEP TO SIMULTE BETA *********************!

        XVX=0.0
        XVY=0.0

        Call OMP_SET_NUM_THREADS(Nthreads)
        !$omp parallel default(shared) private(g,NN, V, Y, YY, &
        !$omp & X, XX, XX2, W, S, V1, V2, zz_1, IPIV, INFO, &
        !$omp & ithread, iblock, ibegin, iend)

        ithread = OMP_GET_THREAD_NUM()
        iblock = (GG+Nthreads-1)/Nthreads
        ibegin = ithread*iblock+1
        iend = min((ithread+1)*iblock,GG)

        DO g = ibegin,iend

            NN=N(g)
            ALLOCATE(V(NN,NN))
            ALLOCATE(Y(NN));        ALLOCATE(YY(NN))
            ALLOCATE(X(NN,15)) ;
            ALLOCATE(XX(NN,30+2*DD)) ; ALLOCATE(XX2(NN,30+2*DD)) ;
            ALLOCATE(W(NN,NN));
            ALLOCATE(S(NN,NN));
            ALLOCATE(V1(NN,NN)); ALLOCATE(V2(NN,NN))
            ALLOCATE(zz_1(NN,DD))
            ALLOCATE(IPIV(NN))

            V=sigma_T(t-1)*EYE(NN)
            Y=group(1:NN,dep,g)
            X=group(1:NN,4:18,g)
            W=network(1:NN,1:NN,g)
            zz_1=zz(1:NN,1:DD,g)
            XX(:,1:15)=X
            XX(:,16:30)=matmul(w_norm(1:NN,1:NN,g),X)
            XX(:,31:30+DD)=zz_1
            XX(:,31+DD:30+2*DD)=matmul(w_norm(1:NN,1:NN,g),zz_1)

            S=eye(NN)-lambda_T(t)*w_norm(1:NN,1:NN,g)

            XX2=XX
            V1=V
            CALL SGESV(NN,30+2*DD,V1,NN,IPIV,XX2,NN,INFO)
            YY=matmul(S,Y)-alpha_T(g,t-1)
            V2=V
            CALL SGESV(NN,1,V2,NN,IPIV,YY,NN,INFO)

            XVX(:,:,g)=matmul(transpose(XX) , XX2)
            XVY(:,g)=matmul(transpose(XX) , YY)

            DEALLOCATE(V)
            DEALLOCATE(Y);    DEALLOCATE(YY)
            DEALLOCATE(X) ;
            DEALLOCATE(XX) ; DEALLOCATE(XX2) ;
            DEALLOCATE(W);
            DEALLOCATE(S);
            DEALLOCATE(V1); DEALLOCATE(V2)
            DEALLOCATE(zz_1)
            DEALLOCATE(IPIV)
        END DO
        !$omp end parallel

        CALL FINDInv(sum(XVX,3)+INV_B_0,INV_B,30+2*DD,ier)
        INV_B2=INV_B
        CALL spotrf('U',30+2*DD, INV_B2, 30+2*DD, info)
        errcode=vsrnggaussianmv(method, stream, 1, beta_temp, 30+2*DD, me, beta_0,  INV_B2)
        beta_T(:,t)=beta_temp+matmul(INV_B,sum(XVY,2))


        !*************GIBBS STEP FOR sigma*********************!

        rho_1=0
        ALLOCATE(ep_v(sum(N)))
        stack=0
        DO g=1,GG
            NN=N(G)
            ALLOCATE(Y(NN))
            ALLOCATE(X(NN,15))
            ALLOCATE(XX(NN,30+2*DD))
            ALLOCATE(W(NN,NN))
            ALLOCATE(S(NN,NN))
            ALLOCATE(ep(NN))
            ALLOCATE(zz_1(NN,DD))

            Y=group(1:NN,dep,g)
            X=group(1:NN,4:18,g)
            W=network(1:NN,1:NN,g)
            zz_1=zz(1:NN,1:DD,g)
            XX(:,1:15)=X
            XX(:,16:30)=matmul(w_norm(1:NN,1:NN,g),X)
            XX(:,31:30+DD)=zz_1
            XX(:,31+DD:30+2*DD)=matmul(w_norm(1:NN,1:NN,g),zz_1)

            S=eye(NN)-lambda_T(t)*w_norm(1:NN,1:NN,g)

            ep=matmul(S,Y)-matmul(XX,beta_T(:,t))-alpha_T(g,t-1)
            ep_v(stack+1:stack+NN)=ep
            rho_1=rho_1+NN
            stack=stack+NN
            DEALLOCATE(Y)
            DEALLOCATE(X)
            DEALLOCATE(XX)
            DEALLOCATE(W)
            DEALLOCATE(S)
            DEALLOCATE(ep)
            DEALLOCATE(zz_1)
        END DO
        sigma_T(t)=(dot_product(ep_v,ep_v)+eta_0)/random_chisq(rho_1+rho_0, first)
        DEALLOCATE(ep_v)

        !*************GIBBS STEP FOR ALPHA_G*********************!


        Call OMP_SET_NUM_THREADS(Nthreads)
        !$omp parallel default(shared) private(g,NN, Y, YY, &
        !$omp & X, XX, W, S, zz_1, IPIV, R_g, &
        !$omp & ithread, iblock, ibegin, iend)

        ithread = OMP_GET_THREAD_NUM()
        iblock = (GG+Nthreads-1)/Nthreads
        ibegin = ithread*iblock+1
        iend = min((ithread+1)*iblock,GG)

        DO g = ibegin,iend

            NN=N(g)
            ALLOCATE(Y(NN));
            ALLOCATE(YY(NN));
            ALLOCATE(X(NN,15))
            ALLOCATE(XX(NN,30+2*DD))
            ALLOCATE(W(NN,NN));
            ALLOCATE(S(NN,NN));
            ALLOCATE(zz_1(NN,DD))

            R_g=(ALPHA_0**(-1)+sigma_T(t)**(-1.0)*NN)**(-1.0)
            Y=group(1:NN,dep,g)
            X=group(1:NN,4:18,g)
            W=network(1:NN,1:NN,g)
            zz_1=zz(1:NN,1:DD,g)
            XX(:,1:15)=X
            XX(:,16:30)=matmul(w_norm(1:NN,1:NN,g),X)
            XX(:,31:30+DD)=zz_1
            XX(:,31+DD:30+2*DD)=matmul(w_norm(1:NN,1:NN,g),zz_1)

            S=eye(NN)-lambda_T(t)*w_norm(1:NN,1:NN,g)

            YY=matmul(S,Y)
            alpha_T(g,t)=R_g*(ALPHA_0**(-1)*a_0+(sigma_T(t))**(-1.0)*sum((YY-matmul(XX,beta_T(:,t)))))+random_normal()*sqrt(R_g)

            DEALLOCATE(Y);
            DEALLOCATE(YY)
            DEALLOCATE(X) ;
            DEALLOCATE(XX) ;
            DEALLOCATE(W);
            DEALLOCATE(S);
            DEALLOCATE(zz_1);
        END DO
        !$omp end parallel


        !!******** Calculate Log likelihood function value *********!

        Call OMP_SET_NUM_THREADS(Nthreads)
        !$omp parallel default(shared) private(g, NN,V, S, Y,X, XX, W, Nu, Co, Ex, &
        !$omp & psi_1, ep_1, ep_1v, INFO, CODE, DD1, FACT, IPVT,  &
        !$omp & C1, C2, C3, V1, zz_1, IPIV, INDX, DET1, DET2, q_1,  &
        !$omp & pp, i, j, like_Y1, rand, &
        !$omp & ithread, iblock, ibegin, iend)

        ithread = OMP_GET_THREAD_NUM()
        iblock = (GG+Nthreads-1)/Nthreads
        ibegin = ithread*iblock+1
        iend = min((ithread+1)*iblock,GG)

        DO g = ibegin,iend

            NN=N(g)
            ALLOCATE(V(NN,NN));        ALLOCATE(S(NN,NN));
            ALLOCATE(Y(NN));
            ALLOCATE(X(NN,15)) ;       ALLOCATE(XX(NN,30+2*DD)) ;
            ALLOCATE(W(NN,NN));
            ALLOCATE(Nu(NN));
            ALLOCATE(Co(NN));          ALLOCATE(Ex(NN));
            ALLOCATE(V1(NN,NN));
            ALLOCATE(psi_1(NN,NN));
            ALLOCATE(ep_1(NN));
            ALLOCATE(ep_1v(NN));
            ALLOCATE(C1(NN,NN));       ALLOCATE(C2(NN,NN));
            ALLOCATE(C3(NN,NN));
            ALLOCATE(zz_1(NN,DD));
            ALLOCATE(IPIV(NN));        ALLOCATE(INDX(NN));
            ALLOCATE(IPVT(NN));        ALLOCATE(FACT(NN,NN));

            V=sigma_T(t)*EYE(NN)
            Y=group(1:NN,dep,g)
            X=group(1:NN,4:18,g)
            Nu=group(1:NN,4,g)
            Co=group(1:NN,5,g)
            Ex=group(1:NN,6,g)
            W=network(1:NN,1:NN,g)
            zz_1=zz(1:NN,1:DD,g)
            XX(:,1:15)=X
            XX(:,16:30)=matmul(w_norm(1:NN,1:NN,g),X)
            XX(:,31:30+DD)=zz_1
            XX(:,31+DD:30+2*DD)=matmul(w_norm(1:NN,1:NN,g),zz_1)

            C1=grade(1:NN,1:NN,g)
            C2=sex(1:NN,1:NN,g)
            C3=race(1:NN,1:NN,g)

            pp=0.0
            DO i=1,NN
                DO j=1, NN
                    IF (j/=i) THEN
                        IF (DD==1) THEN
                            psi_1(i,j)=gamma_T(1,t)+gamma_T(2,t)*c1(i,j)+gamma_T(3,t)*c2(i,j)+gamma_T(4,t)*c3(i,j) &
                                & +gamma_T(5,t)*abs(Nu(i)-Nu(j))+gamma_T(6,t)*abs(Co(i)-Co(j))+gamma_T(7,t)*abs(Ex(i)-Ex(j))-gamma_T(8,t)*abs(zz_1(i,1)-zz_1(j,1))
                        ELSEIF (DD==2) THEN
                            psi_1(i,j)=gamma_T(1,t)+gamma_T(2,t)*c1(i,j)+gamma_T(3,t)*c2(i,j)+gamma_T(4,t)*c3(i,j) &
                                & +gamma_T(5,t)*abs(Nu(i)-Nu(j))+gamma_T(6,t)*abs(Co(i)-Co(j))+gamma_T(7,t)*abs(Ex(i)-Ex(j))-gamma_T(8,t)*abs(zz_1(i,1)-zz_1(j,1))-gamma_T(9,t)*abs(zz_1(i,2)-zz_1(j,2))
                        ELSEIF (DD==3) THEN
                            psi_1(i,j)=gamma_T(1,t)+gamma_T(2,t)*c1(i,j)+gamma_T(3,t)*c2(i,j)+gamma_T(4,t)*c3(i,j) &
                                & +gamma_T(5,t)*abs(Nu(i)-Nu(j))+gamma_T(6,t)*abs(Co(i)-Co(j))+gamma_T(7,t)*abs(Ex(i)-Ex(j))-gamma_T(8,t)*abs(zz_1(i,1)-zz_1(j,1))-gamma_T(9,t)*abs(zz_1(i,2)-zz_1(j,2)) &
                                & -gamma_T(10,t)*abs(zz_1(i,3)-zz_1(j,3))
                        ELSEIF (DD==4) THEN
                            psi_1(i,j)=gamma_T(1,t)+gamma_T(2,t)*c1(i,j)+gamma_T(3,t)*c2(i,j)+gamma_T(4,t)*c3(i,j) &
                                & +gamma_T(5,t)*abs(Nu(i)-Nu(j))+gamma_T(6,t)*abs(Co(i)-Co(j))+gamma_T(7,t)*abs(Ex(i)-Ex(j))-gamma_T(8,t)*abs(zz_1(i,1)-zz_1(j,1))-gamma_T(9,t)*abs(zz_1(i,2)-zz_1(j,2)) &
                                & -gamma_T(10,t)*abs(zz_1(i,3)-zz_1(j,3)) -gamma_T(11,t)*abs(zz_1(i,4)-zz_1(j,4))
                        ELSEIF (DD==5) THEN
                            psi_1(i,j)=gamma_T(1,t)+gamma_T(2,t)*c1(i,j)+gamma_T(3,t)*c2(i,j)+gamma_T(4,t)*c3(i,j) &
                                & +gamma_T(5,t)*abs(Nu(i)-Nu(j))+gamma_T(6,t)*abs(Co(i)-Co(j))+gamma_T(7,t)*abs(Ex(i)-Ex(j))-gamma_T(8,t)*abs(zz_1(i,1)-zz_1(j,1))-gamma_T(9,t)*abs(zz_1(i,2)-zz_1(j,2)) &
                                & -gamma_T(10,t)*abs(zz_1(i,3)-zz_1(j,3)) -gamma_T(11,t)*abs(zz_1(i,4)-zz_1(j,4)) -gamma_T(12,t)*abs(zz_1(i,5)-zz_1(j,5))
                        ELSEIF (DD==6) THEN
                            psi_1(i,j)=gamma_T(1,t)+gamma_T(2,t)*c1(i,j)+gamma_T(3,t)*c2(i,j)+gamma_T(4,t)*c3(i,j) &
                                & +gamma_T(5,t)*abs(Nu(i)-Nu(j))+gamma_T(6,t)*abs(Co(i)-Co(j))+gamma_T(7,t)*abs(Ex(i)-Ex(j))-gamma_T(8,t)*abs(zz_1(i,1)-zz_1(j,1))-gamma_T(9,t)*abs(zz_1(i,2)-zz_1(j,2)) &
                                & -gamma_T(10,t)*abs(zz_1(i,3)-zz_1(j,3)) -gamma_T(11,t)*abs(zz_1(i,4)-zz_1(j,4)) -gamma_T(12,t)*abs(zz_1(i,5)-zz_1(j,5)) -gamma_T(13,t)*abs(zz_1(i,6)-zz_1(j,6))
                        ENDIF
                        q_1=psi_1(i,j)*W(i,j)-log(1+exp(psi_1(i,j)))
                        pp=pp+q_1
                    ENDIF
                END DO
            END DO

            S=eye(NN)-lambda_T(t)*w_norm(1:NN,1:NN,g)

            ep_1=matmul(S,Y)-matmul(XX,beta_T(:,t))-alpha_T(g,t)

            like_Y1=-0.5*NN*log(2*pi)-0.5*NN*log(sigma_T(t))+log(FindDet(S,NN))-dot_product(ep_1, ep_1)/(2*sigma_T(t))

            pp_G(g)=pp+like_Y1

            DEALLOCATE(V);        DEALLOCATE(S);
            DEALLOCATE(Y);
            DEALLOCATE(X) ;       DEALLOCATE(XX);
            DEALLOCATE(Nu);
            DEALLOCATE(Co);       DEALLOCATE(Ex);
            DEALLOCATE(W);
            DEALLOCATE(V1);
            DEALLOCATE(psi_1);
            DEALLOCATE(ep_1);
            DEALLOCATE(ep_1v);
            DEALLOCATE(C1);       DEALLOCATE(C2);
            DEALLOCATE(C3);
            DEALLOCATE(zz_1);
            DEALLOCATE(IPIV);     DEALLOCATE(INDX);
            DEALLOCATE(IPVT);     DEALLOCATE(FACT);
        END DO
        !$omp end parallel

        logp_T(t)=sum(pp_G)


        If(t>burn) THEN
            AICM=2*(var_8(logp_T(burn:t),(t-burn+1))-sum(logp_T(burn:t))/(t-burn+1))
        END IF


        call system_clock ( clock_stop, clock_rate, clock_max )
        t_round=t/50
        IF (t-t_round*50==0) THEN
            WRITE(*,*)   'T=', t
            WRITE(*,'(a6, f7.4)') 'Time:', real ( clock_stop - clock_start ) / real ( clock_rate )
            WRITE(*,'(a6,13f7.4)') 'GAMMA:', gamma_T(:,t)
            WRITE(*,'(a6,f7.4)') 'LAMBDA:', lambda_T(t)
            !WRITE(*,'(a6,8f7.4)') 'BETA:', beta_T(1:8,t)
            !WRITE(*,'(a6,8f7.4)') 'BETA:', beta_T(9:15,t)
            WRITE(*,'(a6, f7.4)') 'SIGMA:',  sigma_T(t)
            WRITE(*,'(a6,f10.2)') 'Likeli:',  logp_T(t)
            WRITE(*,'(a6,f7.4)') 'acc_1:',  acc_rate1(t)
            WRITE(*,'(a6,f7.4)') 'acc_2:',  acc_rate2(t)
            !WRITE(*,'(a6,f7.4)') 'acc_4:',  acc_rate4(t)
            WRITE(*,'(a6,7f7.4)') 'c_0:', c_0(1),c_0(11)
            WRITE(*,'(a6,f15.2)') 'AICM', AICM


            IF (DD==1) THEN
                WRITE(1000,'(8f12.4)')  gamma_T(:,t)
            ELSEIF (DD==2) THEN
                WRITE(1000,'(9f12.4)')  gamma_T(:,t)
            ELSEIF (DD==3) THEN
                WRITE(1000,'(10f12.4)')  gamma_T(:,t)
            ELSEIF (DD==4) THEN
                WRITE(1000,'(11f12.4)')  gamma_T(:,t)
            ELSEIF (DD==5) THEN
                WRITE(1000,'(12f12.4)')  gamma_T(:,t)
            ELSEIF (DD==6) THEN
                WRITE(1000,'(13f12.4)')  gamma_T(:,t)
            ENDIF

            WRITE(2000,'(f10.4)')   lambda_T(t)
            WRITE(3000,'(30f10.4)')  beta_T(:,t)
            WRITE(4000,'(f12.4)')  sigma_T(t)
            !WRITE(6000,100)          alpha_T(:,t)
            WRITE(7000,'(f12.3)')    logp_T(t)
        END IF

    END DO

    END PROGRAM estimation



